from gettext import gettext as _

from gi.repository import Gtk
from gi.repository import GObject

from gdms.enums import BackgroundType
from gdms.utils import GProperty
from gdms.themes import shell_themes, icon_themes, cursor_themes
from gdms.settings import appearance_settings
from gdms.gui.widgets import ImageChooserButton

from .common import PageContent


class AppearancePageContent (PageContent):
    __gtype_name__ = 'AppearancePageContent'

    def __init__ (self, window, **props):
        super().__init__(**props)

        self.window = window

        self.builder = Gtk.Builder.new_from_resource('/app/ui/appearance-page.ui')

        self.set_child(self.builder.get_object('content_box'))

        self.accent_selector = self.builder.get_object('accent_selector')
        self.light_mode_switch = self.builder.get_object('light_mode_switch')

        self.shell_theme_comborow = self.builder.get_object('shell_theme_comborow')
        self.icon_theme_comborow = self.builder.get_object('icon_theme_comborow')
        self.cursor_theme_comborow = self.builder.get_object('cursor_theme_comborow')

        self.background_type_comborow = self.builder.get_object('background_type_comborow')
        self.bg_img_adj_comborow = self.builder.get_object('bg_img_adj_comborow')
        self.background_image_actionrow = self.builder.get_object('background_image_actionrow')
        self.background_color_actionrow = self.builder.get_object('background_color_actionrow')
        self.background_color_button = self.builder.get_object('background_color_button')

        # Add background image button
        self.background_image_button = ImageChooserButton(valign='center')
        self.background_image_actionrow.add_suffix(self.background_image_button)
        self.background_image_actionrow.set_activatable_widget(self.background_image_button)

        self.background_type_comborow.connect('notify::selected', self.on_background_type_changed)

        self.load_theme_lists()
        self.bind_to_gsettings()


    def on_background_type_changed (self, comborow, prop):
        background_type = BackgroundType(comborow.get_selected())
        if background_type is BackgroundType.image:
            self.background_image_actionrow.set_visible(True)
            self.background_color_actionrow.set_visible(False)
        elif background_type is BackgroundType.color:
            self.background_image_actionrow.set_visible(False)
            self.background_color_actionrow.set_visible(True)
        else:
            self.background_image_actionrow.set_visible(False)
            self.background_color_actionrow.set_visible(False)


    def load_theme_lists(self):
        def load_list(comborow, lyst):
            comborow.set_model(Gtk.StringList.new(lyst.names))

        load_list(self.shell_theme_comborow, shell_themes)
        load_list(self.icon_theme_comborow, icon_themes)
        load_list(self.cursor_theme_comborow, cursor_themes)


    def bind_to_gsettings (self):
        appearance_settings.bind('accent-color', self.accent_selector, 'selected')
        appearance_settings.bind('light-mode', self.light_mode_switch, 'active')
        appearance_settings.bind_via_list('shell-theme', self.shell_theme_comborow, 'selected',
                                          shell_themes.theme_ids, strict=False)
        appearance_settings.bind_via_list('icon-theme', self.icon_theme_comborow, 'selected',
                                          icon_themes.theme_ids, strict=False)
        appearance_settings.bind_via_list('cursor-theme', self.cursor_theme_comborow, 'selected',
                                          cursor_themes.theme_ids, strict=False)
        appearance_settings.bind_via_enum('background-type', self.background_type_comborow,
                                          'selected', BackgroundType)
        appearance_settings.bind_via_list('bg-adjustment', self.bg_img_adj_comborow, 'selected',
                                          ['zoom', 'repeat'])
        appearance_settings.bind('background-image', self.background_image_button, 'filename')
        appearance_settings.bind_to_colorbutton('background-color', self.background_color_button)


class AccentButton (Gtk.ToggleButton):
    __gtype_name__ = 'AccentButton'
    accent: str = GProperty(str, construct_only=True)

    def __init__(self, accent: str, **props):
        super().__init__(accent = accent, **props)
        self.add_css_class('accent')
        self.add_css_class(accent)


class AccentSelector (Gtk.Box):
    __gtype_name__ = 'AccentSelector'

    accents = {
        'blue' : _('Blue'),
        'teal' : _('Teal'),
        'green' : _('Green'),
        'yellow' : _('Yellow'),
        'orange' : _('Orange'),
        'red' : _('Red'),
        'pink' : _('Pink'),
        'purple' : _('Purple'),
        'slate' : _('Slate'),
        'maia' : _('Maia'),
    }

    selected: str = GProperty(str, default="blue")
    compact: str = GProperty(bool, default=False)

    def __init__(self, **props):
        super().__init__(**props)
        self.props.halign = Gtk.Align.CENTER
        self.props.valign = Gtk.Align.CENTER

        self.connect("notify::compact", self.on_notify_compact)

        # Add children
        for accent, tooltip in self.accents.items():
            self.append(AccentButton(accent, tooltip_text = tooltip))

        # Set up toggle group
        group = self.get_first_child()
        child = group.get_next_sibling()
        while isinstance(child, AccentButton):
            child.props.group = group
            child = child.get_next_sibling()

        # Connect signals
        child = self.get_first_child()
        while isinstance(child, AccentButton):
            child.connect("toggled", lambda x: x.props.active and self.on_accent_selected(x))
            child = child.get_next_sibling()

        # Update UI when 'selected' property is changed programatically
        self.connect("notify::selected", self.on_notify_selected)
        self.on_notify_selected()

    def on_accent_selected(self, btn: AccentButton):
        self.selected = btn.accent

    def on_notify_compact(self, x=None, y=None):
        if self.compact:
            self.add_css_class("compact")
        else:
            self.remove_css_class("compact")

    def on_notify_selected(self, x=None, y=None):
        child = self.get_first_child()
        while isinstance(child, AccentButton):
            child.props.active = child.accent == self.selected
            child = child.get_next_sibling()



from gettext import gettext as _, pgettext as C_

from gi.repository import Adw

from gdms import VERSION
from gdms import debug_info


mazhar_hussain = C_("Name of Developer", "Mazhar Hussain") + " <realmazharhussain@gmail.com>"
thales_binda   = C_("Name of Artist",    "Thales Bindá") +   " <thales.i.o.b@gmail.com>"


def about_window(win):
    abt = Adw.AboutWindow.new_from_appdata("/app/info.xml", VERSION.split(".")[0] + ".0")
    abt.set_properties(
        artists = [thales_binda],
        copyright = _("Copyright 2021-2023 Mazhar Hussain"),
        debug_info = debug_info.as_markdown(),
        debug_info_filename = 'gdm-settings-debug-info.md',
        developers = [mazhar_hussain],
        documenters = [mazhar_hussain],
        transient_for = win,
        version = VERSION,

        # Translators: Do not translate this string. Put your info here in the form
        # 'name <email>' including < and > but not quotes.
        translator_credits = _("translator-credits"),
    )
    return abt



'''Contains the main Application class'''

import sys
import logging
import subprocess
from configparser import ParsingError
from gettext import gettext as _, pgettext as C_

from gi.repository import Gio, GLib
from gi.repository import Adw, Gtk, Gdk

from gdms import APP_NAME, APP_ID, VERSION, APP_DATA_DIR
from gdms import stderr_log_handler
from gdms import env
from gdms.enums import PackageType
from gdms.utils import BackgroundTask, GSettings
from gdms import gresource
from gdms import debug_info
from gdms import settings

Gio.Resource.load(APP_DATA_DIR + '/resources.gresource')._register()

from .about import about_window
from .window import GdmSettingsWindow

logger = logging.getLogger(__name__)


def set_logging_level(verbosity):
    # Logging Levels are integers with following values
    # logging.CRITICAL = 50
    # logging.ERROR    = 40
    # logging.WARNING  = 30
    # logging.INFO     = 20
    # logging.DEBUG    = 10
    # and integer values above 50 disable logging completely
    # But this is what the values of our app's --verbosity option represent
    # 0 = DISABLE, 1 = CRITICAL, 2 = ERROR, 3 = WARNING, 4 = INFO, 5 = DEBUG
    # So, we use following formula to get appropriate integer that represents chosen level
    # For example, with --verbosity=0 we get, (6-verbosity)*10 = (6-0)*10 = 6*10 = 60 (no logging)
    #            , with --verbosity=1 we get, (6-verbosity)*10 = (6-1)*10 = 5*10 = 50 = logging.CRITICAL
    #         And, with --verbosity=5 we get, (6-verbosity)*10 = (6-5)*10 = 1*10 = 10 = logging.DEBUG
    level = (6 - verbosity) * 10
    stderr_log_handler.setLevel(level)


class GdmSettingsApp(Adw.Application):
    '''The main Application class'''

    __gtype_name__ = "GdmSettingsApp"

    def __init__(self):
        super().__init__(application_id=APP_ID)
        self.props.resource_base_path = '/app/'

        def add_option(long_name, short_name, description, argument_name=None, argument_type='NONE'):
            self.add_main_option(long_name, ord(short_name),
                                 GLib.OptionFlags.NONE,
                                 getattr(GLib.OptionArg, argument_type),
                                 description, argument_name,
                                 )

        add_option('debug-info', 'i',  _('Show debug information (in Markdown format)'))
        add_option('version',    '\0', _('Show application version'))
        add_option('verbosity',  '\0', _('Set verbosity level manually (from 0 to 5)'),
                   C_('Argument of --verbosity option', 'LEVEL'), 'INT')
        add_option('verbose',    'v',  _('Enable verbose mode (alias of --verbosity=5)'))
        add_option('quiet',      'q',  _('Enable quiet mode (alias of --verbosity=0)'))


    def do_handle_local_options(self, options):
        if options.contains("version"):
            print(APP_NAME, VERSION)
            return 0

        if options.contains("debug-info"):
            print(debug_info.as_markdown())
            return 0

        set_logging_level(3)

        if options.contains("verbose"):
            set_logging_level (5)

        if options.contains("quiet"):
            set_logging_level (0)

        if verbosity_gvariant := options.lookup_value("verbosity", GLib.VariantType("i")):
            verbosity_level = verbosity_gvariant.get_int32()

            if verbosity_level < 0 or verbosity_level > 5:
                print(_('Invalid verbosity level {level}. Must be 0, 5, or in-between.'
                       ).format(level=verbosity_level),
                      file=sys.stderr)
                return 1

            set_logging_level(verbosity_level)

        return -1


    def do_startup(self):
        Adw.Application.do_startup(self)
        settings.init()

    def do_activate(self):
        if win := self.get_active_window():
            win.present()
            return

        BOLD = '\033[1m'
        NORMAL = '\033[0m'
        for key, val in debug_info.debug_info:
            logger.info(f"{BOLD}{key}{NORMAL}: {val}")

        self.settings = GSettings(APP_ID)

        self.reset_settings_task = BackgroundTask(settings.reset,
                                                  self.on_reset_settings_finish)

        self.import_task = BackgroundTask(None, self.on_import_finished)
        self.export_task = BackgroundTask(None, self.on_export_finished)

        self.create_actions()
        self.keyboard_shortcuts()

        self.window = GdmSettingsWindow(self)
        self.window.present()

        # Some dependencies cannot be packaged along with the app but are
        # required to be installed on the user's system. So, we need to
        # check them and report to the user if they are missing.
        if not self.check_system_dependencies():
            return

        if not gresource.get_default():
            self.show_missing_gresource_dialog()
            return

        if self.settings['initial-warning']:
            self.show_initial_warning()

        if (not self.settings['donation-dialog-shown']
        and not self.settings['never-applied']):
            self.show_donation_dialog()


    def do_shutdown(self):
        settings.finalize()
        Adw.Application.do_shutdown(self)


    def check_system_dependencies(self):
        '''If some dependencies are missing, show a dialog reporting the situation'''

        def check_dependency(exec_name, logging_name=None, *, on_host=True):
            host_args = []
            if env.PACKAGE_TYPE is PackageType.Flatpak and on_host is True:
                host_args = ['flatpak-spawn', '--host']

            try:
                proc = subprocess.run([*host_args, exec_name, '--version'], capture_output=True)
                if proc.returncode == 0:
                    version_info = proc.stdout.decode().strip()
                    if logging_name:
                        logger.info(f'{logging_name} {version_info}')
                    else:
                        logger.info(version_info)
                    return True
            except FileNotFoundError: pass

            return False

        gdm_installed = bool(gresource.ShellGresourceFile)
        polkit_installed = check_dependency('pkexec')

        if gdm_installed and polkit_installed:
            return True

        if env.PACKAGE_TYPE is PackageType.Flatpak:
            message = _('Following programs are required to be installed <b>on the host system</b> for'
                        ' this app to function properly but they are not installed on the host system.'
                       )
        else:
            message = _('This app requires the following software to function properly but'
                        ' they are not installed.'
                       )

        message += '\n'

        if not gdm_installed:
            message += '\n'
            message += C_('Missing Dependency',
                          '<b>GDM</b>'
                         )

        if not polkit_installed:
            message +='\n'
            message += C_('Missing Dependency',
                          '<b>Polkit</b>'
                         )

        message += '\n\n'

        link = 'https://github.com/gdm-settings/gdm-settings/wiki/Dependencies#how-to-install-dependencies'
        # Translators: Keep '<a href="{url}">' and '</a>' as is. The words between them will become
        # a link to '{url}' and '{url}' will be replaced by a real URL during program execution.
        message += _('Click <a href="{url}">here</a> for instructions on how '
                     'to install these dependencies on your system.'
                    ).format(url=link)

        dialog = Adw.MessageDialog(
                    body = message,
                   modal = True,
                 heading = _('Missing Dependencies'),
           transient_for = self.window,
         body_use_markup = True,
        )

        dialog.add_response('ok', _('OK'))
        dialog.connect('response', lambda *args: self.quit())
        dialog.present()


    def show_missing_gresource_dialog (self):
        heading = _("Default Shell Theme File Lost")
        body = _(
            "The app needs but could not find an unmodified default shell theme file. "
            "To fix the issue, reinstall 'gnome-shell-common' or 'gnome-shell' package."
        )

        dialog = Adw.MessageDialog.new(self.window, heading, body)
        dialog.add_response('ok', _('OK'))
        dialog.connect('response', lambda *args: self.quit())
        dialog.present()


    def show_initial_warning (self):
        heading = _("Caution!")
        body = _(
            "In order for some important features to work, the app modifies a critical system file"
            " and has the potential to <b>break your system</b>."
        )

        check_button = Gtk.CheckButton(label=_("Don't show again"), halign=Gtk.Align.CENTER)
        self.settings.bind("initial-warning", check_button, "active", Gio.SettingsBindFlags.INVERT_BOOLEAN)

        dialog = Adw.AlertDialog.new(heading, body)
        dialog.props.body_use_markup = True
        dialog.props.extra_child = check_button
        dialog.add_response('ok', _("I Understand"))
        dialog.present(self.window)


    def show_donation_dialog (self):
        heading = _("Donation Request")
        body = _(
            "If you like this app, please consider donating so that we can keep improving it.\n"
            "Thank you! ❤️️"
        )

        dialog = Adw.AlertDialog.new(heading, body)

        dialog.add_response('close', _("Not Interested"))
        dialog.add_response('donate', _("Donate"))

        dialog.set_response_appearance('donate', Adw.ResponseAppearance.SUGGESTED)
        dialog.set_default_response('donate')

        dialog.connect('response::donate', lambda *args: self.activate_action('donate'))

        dialog.present(self.window)

        self.settings['donation-dialog-shown'] = True


    def keyboard_shortcuts(self):
        self.set_accels_for_action("app.quit", ["<Ctrl>q"])
        self.set_accels_for_action("app.refresh", ["<Ctrl>r", "F5"])


    def create_actions(self):

        def create_action(action_name, function):
            action = Gio.SimpleAction(name=action_name)
            action.connect("activate", function)
            self.add_action(action)

        create_action("refresh", self.refresh_cb)
        create_action("load_session_settings", self.load_session_settings_cb)
        create_action("reset_settings", self.reset_settings_cb)
        create_action("import_from_file", self.import_from_file_cb)
        create_action("export_to_file", self.export_to_file_cb)
        create_action("donate", self.donate_cb)
        create_action("about", self.about_cb)
        create_action("quit", self.quit_cb)


    def refresh_cb(self, action, user_data):
        settings.drop_unapplied_changes()

        toast = Adw.Toast(timeout=1, priority='high', title=_('Settings reloaded'))
        self.window.toast_overlay.add_toast(toast)


    def load_session_settings_cb(self, action, user_data):
        if env.PACKAGE_TYPE is not PackageType.Flatpak:
            settings.load_from_session()
            toast = Adw.Toast(timeout=1, priority='high', title=_('Session settings loaded successfully'))
            self.window.toast_overlay.add_toast(toast)
        else:
            toast = Adw.Toast(timeout=2, priority="high")
            toast.set_title(_("Load session settings is NOT supported in Flatpak version"))
            self.window.toast_overlay.add_toast(toast)


    def reset_settings_cb(self, action, user_data):
        self.window.task_counter.inc()
        self.reset_settings_task.start()

    def on_reset_settings_finish(self):
        success = self.reset_settings_task.finish()
        self.window.task_counter.dec()

        if success:
            message = _("Reset settings successfully")
        else:
            message = _("Failed to reset settings")

        toast = Adw.Toast(timeout=2, priority="high", title=message)
        self.window.toast_overlay.add_toast(toast)


    def import_from_file_cb(self, action, user_data):

        def on_file_chooser_response(file_chooser, response):
            if response == Gtk.ResponseType.ACCEPT:
                self.window.task_counter.inc()
                filepath = file_chooser.get_file().get_path()
                self.import_task.function = lambda: settings.load_from_file(filepath)
                self.import_task.start()
            file_chooser.destroy()

        self._file_chooser = Gtk.FileChooserNative(
                                      modal = True,
                                     action = Gtk.FileChooserAction.OPEN,
                               accept_label = _('Import'),
                              transient_for = self.window,
                             )

        ini_filter = Gtk.FileFilter(name='INI Files')
        ini_filter.add_suffix('ini')
        self._file_chooser.add_filter(ini_filter)
        self._file_chooser.set_filter(ini_filter)

        all_filter = Gtk.FileFilter(name='All Files')
        all_filter.add_pattern('*')
        self._file_chooser.add_filter(all_filter)

        self._file_chooser.connect('response', on_file_chooser_response)
        self._file_chooser.show()

    def on_import_finished(self):
        self.window.task_counter.dec()
        try:
            self.import_task.finish()

            message = _('Settings were successfully imported from file')
        except (ParsingError, UnicodeDecodeError):
            message = _('Failed to import. File is invalid')

        toast = Adw.Toast(timeout=2, priority="high", title=message)
        self.window.toast_overlay.add_toast(toast)


    def export_to_file_cb(self, action, user_data):

        def on_file_chooser_response(file_chooser, response):
            if response == Gtk.ResponseType.ACCEPT:
                self.window.task_counter.inc()
                filepath = file_chooser.get_file().get_path()
                self.export_task.function = lambda: settings.save_to_file(filepath)
                self.export_task.start()
            file_chooser.destroy()

        self._file_chooser = Gtk.FileChooserNative(
                                      modal = True,
                                     action = Gtk.FileChooserAction.SAVE,
                              transient_for = self.window,
                             )

        self._file_chooser.set_current_name('gdm-settings.ini')
        self._file_chooser.connect('response', on_file_chooser_response)
        self._file_chooser.show()

    def on_export_finished(self):
        self.window.task_counter.dec()
        try:
            self.export_task.finish()

            message = _('Settings were successfully exported')
        except PermissionError:
            message = _('Failed to export. Permission denied')
        except IsADirectoryError:
            message = _('Failed to export. A directory with that name already exists')

        toast = Adw.Toast(timeout=2, priority="high", title=message)
        self.window.toast_overlay.add_toast(toast)


    def donate_cb(self, action, user_data):
        Gtk.show_uri(self.window, 'https://opencollective.com/gdm-settings', Gdk.CURRENT_TIME)


    def about_cb(self, action, user_data):
        win = about_window(self.window)
        win.present()


    def quit_cb(self, action, user_data):
        self.quit()



'''Python package for GDM Settings app'''

import os
import sys
import gettext
import locale
import logging
import tempfile
from enum import Enum

from gettext import gettext as _

import gi
gi.require_version('Adw', '1')

from . import config

# This app is relocatable. APP_DIR (detected during runtime) is path
# to the base directory in which this app is running from.
# For example, /opt/gdm-settings, /tmp/.mount_GDM_Se1tvOjd, etc.
APP_DIR = os.path.realpath(__file__).split(config.prefix)[0]

_localedir = APP_DIR + config.localedir
locale.bindtextdomain('gdm-settings', _localedir)
locale.textdomain('gdm-settings')
gettext.bindtextdomain('gdm-settings', _localedir)
gettext.textdomain('gdm-settings')

# Prefer the data dir where the app is installed over other data dirs
_data_dirs = os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share').split(':')
_app_data_dir = APP_DIR + config.datadir
if _app_data_dir in _data_dirs:
    _data_dirs.remove(_app_data_dir)
os.environ['XDG_DATA_DIRS'] = ':'.join([_app_data_dir, *_data_dirs])

APP_ID = 'io.github.realmazharhussain.GdmSettings'
APP_NAME = _('GDM Settings')
PROJECT_NAME = 'gdm-settings'

VERSION = config.version
DEBUG = config.debug
APP_DATA_DIR = APP_DIR + config.datadir + '/' + PROJECT_NAME


# Set up logging

class _Style(str, Enum):
    RED = '\033[31m'
    BOLD = '\033[1m'
    BLUE = '\033[34m'
    GREEN = '\033[32m'
    NORMAL = '\033[0m'
    YELLOW = '\033[33m'
    MANENTA = '\033[35m'
    BRIGHT_RED = '\033[91m'


class _StdErrFormatter(logging.Formatter):
    def format (self, record):
        match record.levelname:
            case 'CRITICAL': level_color = _Style.BRIGHT_RED
            case 'ERROR':    level_color = _Style.RED
            case 'WARNING':  level_color = _Style.YELLOW
            case 'INFO':     level_color = _Style.GREEN
            case default:    level_color = _Style.BLUE

        return (_Style.BOLD + level_color + record.levelname + _Style.NORMAL + ':'
                + _Style.MANENTA + record.name + _Style.NORMAL + ':'
                + ' ' + record.getMessage())


logger = logging.getLogger(__name__)

logger.setLevel(logging.DEBUG)

# FIXME: logging level of handlers can be changed only
# if we set it to a high value here

stderr_log_handler = logging.StreamHandler()
stderr_log_handler.setFormatter(_StdErrFormatter())
stderr_log_handler.setLevel(logging.DEBUG)
logger.addHandler(stderr_log_handler)


def main() -> int:
    from .gui import GdmSettingsApp
    gui = GdmSettingsApp()
    
    try:
        return gui.run(sys.argv)
    except KeyboardInterrupt:
        return 0



'''Enums used in the app'''

from enum import Enum, auto


class PackageType (Enum):
    Unknown  = auto()
    AppImage = auto()
    Flatpak  = auto()


class BackgroundType (Enum):
    default = 0
    image   = 1
    color   = 2



'''A collection of general purpose and reusable widgets'''

from gettext import gettext as _

from gi.repository import Adw
from gi.repository import Gtk
from gi.repository import Pango
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import GLib

from gdms.utils import GProperty


class FileChooserButton (Gtk.Button):
    __gtype_name__ = 'FileChooserButton'

    title = GProperty(str, default=_('Choose File'))
    filters = GProperty(Gio.ListModel)
    default_filter = GProperty(Gtk.FileFilter)


    def __init__ (self, **props):

        none_label = Gtk.Label(label=_('(None)'))

        self._file_image = Gtk.Image()
        self._file_label = Gtk.Label(ellipsize=Pango.EllipsizeMode.END, hexpand=True)

        file_box = Gtk.Box(orientation='horizontal', spacing=5)
        file_box.append(self._file_image)
        file_box.append(self._file_label)

        self._stack = Gtk.Stack(halign=Gtk.Align.START)
        self._stack.add_named(none_label, 'none_label')
        self._stack.add_named(file_box, 'file_box')
        self._stack.set_visible_child_name('none_label')

        main_box = Gtk.Box(orientation='horizontal', spacing=5)
        main_box.append(self._stack)
        main_box.append(Gtk.Separator(css_classes=['spacer'], halign=Gtk.Align.END, hexpand=True))
        main_box.append(Gtk.Image(icon_name='document-open-symbolic', halign=Gtk.Align.END))

        super().__init__(child=main_box, **props)

        self.file_dialog = Gtk.FileDialog(
            modal = True,
            title = self.title,
            filters = self.filters,
            default_filter = self.default_filter,
            accept_label = _('Choose'))


    @GProperty(str, default='')
    def filename (self):
        return self.file.get_path()

    @filename.setter
    def filename (self, value):
        self.file = Gio.File.new_for_path(value)


    @GProperty(Gio.File)
    def file (self):
        return self._file

    @file.setter
    def file (self, value):
        self._file = value
        self.notify('filename')
        self._update_ui()


    def _update_ui (self):
        if not self.file.get_path():
            self._stack.set_visible_child_name('none_label')
            return

        if self.file.query_exists(None):
            file_info = self.file.query_info('standard::*', 0, None)
            self._file_image.props.gicon = file_info.get_icon()
            self._file_label.props.label = file_info.get_display_name()
        else:
            self._file_image.props.icon_name = 'warning'
            self._file_label.props.label = self.file.get_basename()

        self._stack.set_visible_child_name('file_box')


    def do_clicked (self):
        self.file_dialog.set_initial_file(self.file),
        self.file_dialog.open(
            parent = self.get_root(),
            callback = self.file_dialog_open_finish_cb)

    def file_dialog_open_finish_cb(self, file_dialog, result):
        try:
            self.file = file_dialog.open_finish(result)
        except GLib.Error as err:
            if not err.matches(Gtk.dialog_error_quark(), Gtk.DialogError.DISMISSED):
                raise


class ImageChooserButton (FileChooserButton):
    __gtype_name__ = 'ImageChooserButton'

    def __init__ (self, **props):
        all_filters = Gio.ListStore()

        image_filter = Gtk.FileFilter(name=_('Images'))
        image_filter.add_mime_type('image/*')
        all_filters.append(image_filter)

        all_files_filter = Gtk.FileFilter(name=_('All Files'))
        all_files_filter.add_pattern('*')
        all_filters.append(all_files_filter)

        super().__init__(
            title = _('Choose Image'),
            filters = all_filters,
            default_filter = image_filter,
            **props)



from gi.repository import Gtk

from gdms.settings import power_settings

from .common import PageContent


class PowerPageContent (PageContent):
    __gtype_name__ = 'PowerPageContent'

    def __init__ (self, window, **props):
        super().__init__(**props)

        self.window = window

        self.builder = Gtk.Builder.new_from_resource('/app/ui/power-page.ui')

        self.set_child(self.builder.get_object('content_box'))

        self.power_button_comborow = self.builder.get_object('power_button_comborow')
        self.auto_power_saver_row = self.builder.get_object('auto_power_saver_row')
        self.dim_screen_row = self.builder.get_object('dim_screen_row')
        self.screen_blank_switch = self.builder.get_object('screen_blank_switch')
        self.screen_blank_spinrow = self.builder.get_object('screen_blank_spinrow')
        self.suspend_on_battery_switch = self.builder.get_object('suspend_on_battery_switch')
        self.suspend_on_battery_spinrow = self.builder.get_object('suspend_on_battery_spinrow')
        self.suspend_on_ac_switch = self.builder.get_object('suspend_on_ac_switch')
        self.suspend_on_ac_spinrow = self.builder.get_object('suspend_on_ac_spinrow')

        self.bind_to_gsettings()

    def bind_to_gsettings (self):
        power_settings.bind_via_list('power-button-action', self.power_button_comborow, 'selected',
                                     ['nothing', 'suspend', 'hibernate', 'interactive'])
        power_settings.bind('auto-power-saver', self.auto_power_saver_row, 'active')
        power_settings.bind('dim-screen', self.dim_screen_row, 'active')
        power_settings.bind('blank-screen', self.screen_blank_switch, 'active')
        power_settings.bind('idle-delay', self.screen_blank_spinrow, 'value')
        power_settings.bind('suspend-on-battery', self.suspend_on_battery_switch, 'active')
        power_settings.bind('suspend-on-battery-delay', self.suspend_on_battery_spinrow, 'value')
        power_settings.bind('suspend-on-ac', self.suspend_on_ac_switch, 'active')
        power_settings.bind('suspend-on-ac-delay', self.suspend_on_ac_spinrow, 'value')



from gi.repository import Adw
from gi.repository import Gtk


class PageContent (Gtk.ScrolledWindow):
    '''Base class for content of a page from Gtk.Stack'''
    def __init__ (self, **props):
        super().__init__(**props)

        self.set_propagate_natural_height(True)

        margin=20
        self.clamp = Adw.Clamp(maximum_size=1500, tightening_threshold=0)
        self.clamp.set_margin_top(margin)
        self.clamp.set_margin_bottom(margin)
        self.clamp.set_margin_start(margin)
        self.clamp.set_margin_end(margin)
        super().set_child(self.clamp)

    def set_child (self, child):
        self.clamp.set_child(child)



from typing import cast

from gi.repository import Adw
from gi.repository import Gtk

from gdms.settings import font_settings
from .common import PageContent


class FontsPageContent (PageContent):
    __gtype_name__ = 'FontsPageContent'

    def __init__ (self, window, **props):
        super().__init__(**props)

        self.window = window

        self.builder = Gtk.Builder.new_from_resource('/app/ui/fonts-page.ui')

        self.set_child(cast(Gtk.Widget, self.builder.get_object('content_box')))

        self.font_button = cast(Gtk.FontDialogButton, self.builder.get_object('font_button'))
        self.antialiasing_comborow = cast(Adw.ComboRow, self.builder.get_object('antialiasing_comborow'))
        self.hinting_comborow = cast(Adw.ComboRow, self.builder.get_object('hinting_comborow'))
        self.scaling_factor_spinrow = cast(Adw.SpinRow, self.builder.get_object('scaling_factor_spinrow'))

        self.bind_to_gsettings()

    def bind_to_gsettings (self):
        font_settings.bind_to_fontbutton('font', self.font_button)
        font_settings.bind_via_list('antialiasing', self.antialiasing_comborow, 'selected',
                                    ['grayscale', 'rgba', 'none'])
        font_settings.bind_via_list('hinting', self.hinting_comborow, 'selected',
                                    ['full', 'medium', 'slight', 'none'])
        font_settings.bind('scaling-factor', self.scaling_factor_spinrow, 'value')



import os
import subprocess
from collections.abc import Sequence


from gdms.env import TEMP_DIR, PACKAGE_TYPE
from gdms.enums import PackageType


def _unpack_args(args: tuple):
    # Validate Args
    if len(args) == 0:
        raise ValueError('No argument provided')

    # Unpack Args
    if len(args) == 1:
        return args[0]
    
    return args


class CommandResult(subprocess.CompletedProcess):
    '''Result of a command's execution
    
    Returned by Command.run()
    
    Allows usage patterns like

        if Command('echo "Hello, World!"').run():
            print("Mission accomplished!")

        if Command('which non-existant-program').run().failure:
            print("'non-existant-program' does not exist!")
    '''

    def __init__(self, other: subprocess.CompletedProcess, /):
        super().__init__(args=other.args, returncode=other.returncode,
                         stdout=other.stdout, stderr=other.stderr)

    @property
    def success(self):
        return self.returncode == 0

    @property
    def failure(self):
        return not self.success

    def __bool__ (self):
        return self.success


class Command:
    '''A command-line command'''

    __slots__ = ['_data']
    __dict__ = None

    def __new__(cls, *args: tuple[str] | str):
        args = _unpack_args(args)

        # Create instance
        self = object.__new__(cls)

        # Initialize instance
        if isinstance(args, str):
            self._data = args
        else:
            self._data = [str(arg) for arg in args]

        return self
    
    def __str__(self) -> str:
        if isinstance(self._data, str):
            return self._data
        return ' '.join(self._data)

    def prefix(self, *prefix: str):
        command = self._data
        if isinstance(command, str):
            command = ' '.join([*prefix, command])
        else:
            command = [*prefix, *command]
        return type(self)(command)

    def _run(self, *args, **kwargs):
        proc = subprocess.run(self._data, *args, shell=isinstance(self._data, str), **kwargs)
        return CommandResult(proc)

    def run(self, *args, as_root=False, escape_sandbox=True, **kwargs):
        command = self

        if as_root:
            command = command.prefix('pkexec')

        if escape_sandbox and PACKAGE_TYPE is PackageType.Flatpak:
            command = command.prefix('flatpak-spawn', '--host')

        return command._run(*args, **kwargs)


class CommandList:
    """A list of commands"""

    def __init__(self, *, initial_list: Sequence[Command] = []):
        self._list = list(initial_list)
    
    def __str__(self):
        return '\n'.join(str(command) for command in self._list)

    @property
    def empty (self):
        return len(self._list) == 0

    def add(self, *command: Command | tuple[str] | str) -> Command:
        """ Add a new command to the list """
        command = _unpack_args(command)

        if not isinstance(command, Command):
            command = Command(command)

        self._list.append(command)

        return command

    def clear(self):
        """ Clear command list """
        self._list.clear()

    def save_to_file(self, filepath: str):
        filedir = os.path.dirname(filepath)
        os.makedirs(filedir, exist_ok=True)

        with open(filepath, "w") as f:
            f.write(str(self))

    def run(self, *args, as_root=True, stop_on_error=True, clear_on_finish=True, **kwargs):
        '''Run commands as root and clear the list'''

        script_path = f"{TEMP_DIR}/run-elevated"
        self.save_to_file(script_path)

        sh_args = []
        if stop_on_error:
            sh_args.append('-e')

        proc = Command('sh', *sh_args, script_path).run(*args, as_root=as_root, **kwargs)
        os.remove(script_path)

        if clear_on_finish:
            self.clear()

        return proc



'''get information about program's environment'''

# This is an initializer module i.e. it is a module that is assumed to be
# loaded very early in the program launch process and other modules (that
# are not self-contained) can depend on it.
#
# So, this module should not depend on any other module from the gdm_settings
# package unless the other module is fully self-contained.

import os
import re
import ast
import sys

from gi.repository import GLib

from gdms import APP_ID
from gdms.enums import PackageType


def get_os_info():
    filename = None
    for fn in '/run/host/os-release', '/etc/os-release', '/usr/lib/os-release':
        if os.path.isfile(fn):
            filename = fn
            break

    if filename is None:
        return

    os_release = []
    with open(filename, 'r') as file:
        for line_number, line in enumerate(file, start=1):
            line = line.strip()         # Strip whitespace

            if not line or line.startswith('#'):
                continue

            if m := re.match(r'([A-Z][A-Z_0-9]+)=(.*)', line):
                name, val = m.groups()
                if val and val[0] in '"\'':
                    val = ast.literal_eval(val)
                os_release.append((name, val))
            else:
                print(f'{filename}:{line_number}: bad line {line!r}',
                      file=sys.stderr)

    return dict(os_release)

class PATH (list):
    '''
    A list to store values of PATH-like environment variables

    For example, with
        mypath = PATH('/usr/local/bin:/usr/bin')
    OR
        mypath = PATH(['/usr/local/bin', '/usr/bin'])
    we get,
        print(mypath)       # prints /usr/local/bin:/usr/bin
        print(*mypath)      # prints /usr/local/bin /usr/bin
        print(mypath[0])    # prints /usr/local/bin
        print(repr(mypath)) # prints PATH(['/usr/local/bin', '/usr/bin'])
    '''

    def __init__ (self, value=None, /):
        if value is None:
            return list.__init__(self)
        elif isinstance(value, str):
            value = value.strip(':').split(':')
        return list.__init__(self, value)

    def __str__ (self, /):
        return ':'.join(self)

    def __repr__ (self, /):
        return self.__class__.__name__ + '(' + list.__repr__(self) + ')'

    def copy (self, /):
        return self.__class__(self)


# XDG Base Directories
XDG_CONFIG_HOME = GLib.get_user_config_dir()
XDG_RUNTIME_DIR = GLib.get_user_runtime_dir()


# Application-specific Directories
TEMP_DIR       = os.path.join(XDG_RUNTIME_DIR, 'app', APP_ID)
HOST_DATA_DIRS = PATH(os.environ.get('HOST_DATA_DIRS', '/usr/local/share:/usr/share'))


# Package Type and related stuff
PACKAGE_TYPE = PackageType.Unknown
HOST_ROOT    = ''

if os.environ.get('FLATPAK_ID'): # Flatpak
    PACKAGE_TYPE = PackageType.Flatpak
    HOST_ROOT    = '/run/host'
elif os.environ.get('APPDIR'):   # AppImage
    PACKAGE_TYPE = PackageType.AppImage


# OS Release info
os_info = get_os_info()

OS_NAME       = os_info.get('NAME',       'Linux')
OS_ID         = os_info.get('ID',         'linux')
OS_ID_LIKE    = os_info.get('ID_LIKE',    'linux')
OS_VERSION_ID = os_info.get('VERSION_ID', '0')
OS_VERSION    = os_info.get('VERSION', OS_VERSION_ID)

OS_VERSION_CODENAME = os_info.get('VERSION_CODENAME')
_pretty_name = f'{OS_NAME} {OS_VERSION}'
if OS_VERSION_CODENAME: _pretty_name += f' ({OS_VERSION_CODENAME})'

OS_PRETTY_NAME = os_info.get('PRETTY_NAME', _pretty_name)

OS_IDs = [OS_ID] + OS_ID_LIKE.split()



from gi.repository import Adw
from gi.repository import Gtk
from gi.repository import GObject

from gdms.utils import GProperty


class SidebarItem(Adw.Bin):
    icon_name = GProperty(str)
    label = GProperty(str)
    clicked = GObject.Signal()

    def __init__(self, **props):
        super().__init__(**props)

        box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 10)

        icon = Gtk.Image.new()
        self.bind_property("icon-name", icon, "icon-name")
        box.append(icon)

        label = Gtk.Label.new()
        self.bind_property("label", label, "label")
        box.append(label)

        click = Gtk.GestureClick.new()
        click.connect("released", self.on_click)

        self.props.child = box
        self.props.height_request = 40
        self.add_controller(click)

    def on_click(self, ctrl, n_press, x, y):
        self.emit("clicked")


class Sidebar(Adw.Bin):
    __gtype_name__ = 'GdmSettingsSidebar'

    stack = GProperty(Gtk.Stack)
    activate = GObject.Signal()

    def __init__(self):
        super().__init__()

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self.setup_list_item)
        factory.connect("bind", self.bind_list_item)

        self.listview = Gtk.ListView.new(None, factory)
        self.listview.connect("activate", lambda x,y: self.emit("activate"))
        self.listview.add_css_class("navigation-sidebar")

        self.props.child = Gtk.ScrolledWindow(child=self.listview)
        self.connect("notify::stack", self.notify_stack_cb)

    @staticmethod
    def notify_stack_cb (this, prop):
        if this.stack:
            pages = this.stack.props.pages
            this.listview.props.model = pages
        else:
            this.listview.props.model = None

    def setup_list_item (self, factory, item: Gtk.ListItem):
        s_item = SidebarItem()
        s_item.connect("clicked", lambda x: self.emit("activate"))
        item.props.child = s_item

    def bind_list_item (self, factory, item: Gtk.ListItem):
        page: Gtk.StackPage = item.props.item
        s_item: SidebarItem = item.props.child

        page.bind_property("icon-name", s_item, "icon-name", GObject.BindingFlags.SYNC_CREATE)
        page.bind_property("title", s_item, "label", GObject.BindingFlags.SYNC_CREATE)



import os
import shutil
import logging
import sys
from math import trunc
from configparser import ConfigParser, ParsingError
from gettext import gettext as _, pgettext as C_

from gi.repository import Gio

from gdms import APP_ID
from gdms import env
from gdms import gresource
from gdms.cmd import CommandList
from gdms.enums import PackageType, BackgroundType
from gdms.utils import GSettings
from gdms.themes import shell_themes

logger = logging.getLogger(__name__)


main_settings          = GSettings.new_delayed(APP_ID)
accessibility_settings = GSettings.new_delayed(APP_ID + '.accessibility')
appearance_settings    = GSettings.new_delayed(APP_ID + '.appearance')
font_settings          = GSettings.new_delayed(APP_ID + '.fonts')
login_screen_settings  = GSettings.new_delayed(APP_ID + '.misc')
night_light_settings   = GSettings.new_delayed(APP_ID + '.night-light')
mouse_settings         = GSettings.new_delayed(APP_ID + '.mouse')
pointing_settings      = GSettings.new_delayed(APP_ID + '.pointing')
power_settings         = GSettings.new_delayed(APP_ID + '.power')
touchpad_settings      = GSettings.new_delayed(APP_ID + '.touchpad')
sound_settings         = GSettings.new_delayed(APP_ID + '.sound')
top_bar_settings       = GSettings.new_delayed(APP_ID + '.top-bar')


all_settings = (
    main_settings,
    accessibility_settings,
    appearance_settings,
    font_settings,
    login_screen_settings,
    night_light_settings,
    mouse_settings,
    pointing_settings,
    power_settings,
    touchpad_settings,
    sound_settings,
    top_bar_settings,
)


class LogoImageNotFoundError (FileNotFoundError): pass


def _GSettings(schema_id):
    if Gio.SettingsSchemaSource.get_default().lookup(schema_id, recursive=True):
        return GSettings(schema_id)


_commands = CommandList()


def init():
    '''Initialize the settings module'''
    os.makedirs(env.TEMP_DIR, exist_ok=True)
    if main_settings["never-applied"] and env.PACKAGE_TYPE is not PackageType.Flatpak:
        load_from_session()


def finalize():
    '''Finalize the settings module'''
    shutil.rmtree(path=env.TEMP_DIR, ignore_errors=True)


def load_from_session():
    '''Load user's Gnome settings into the app'''

    if user_settings := _GSettings('org.gnome.shell.extensions.user-theme'):
        appearance_settings['shell-theme'] = user_settings['name']

    if user_settings := _GSettings("org.gnome.desktop.a11y"):
        source_key = "always-show-universal-access-status"
        target_key = "always-show-accessibility-menu"
        accessibility_settings[target_key] = user_settings[source_key]

    if user_settings := _GSettings("org.gnome.desktop.interface"):
        appearance_settings['light-mode'] = user_settings["color-scheme"] == 'prefer-light'
        appearance_settings['accent-color'] = user_settings["accent-color"]
        appearance_settings['icon-theme'] = user_settings["icon-theme"]
        appearance_settings['cursor-theme'] = user_settings["cursor-theme"]

        font_settings['font'] = user_settings["font-name"]
        font_settings['antialiasing'] = user_settings["font-antialiasing"]
        font_settings['hinting'] = user_settings["font-hinting"]
        font_settings['scaling-factor'] = user_settings["text-scaling-factor"]

        pointing_settings['cursor-size'] = user_settings["cursor-size"]

        top_bar_settings['show-date'] = user_settings["clock-show-date"]
        top_bar_settings['show-weekday'] = user_settings["clock-show-weekday"]
        top_bar_settings['time-format'] = user_settings["clock-format"]
        top_bar_settings['show-seconds'] = user_settings["clock-show-seconds"]
        top_bar_settings['show-battery-percentage'] = user_settings["show-battery-percentage"]

    if user_settings := _GSettings("org.gnome.desktop.sound"):
        sound_settings['theme'] = user_settings["theme-name"]
        sound_settings['event-sounds'] = user_settings["event-sounds"]
        sound_settings['feedback-sounds'] = user_settings["input-feedback-sounds"]
        sound_settings['over-amplification'] = user_settings["allow-volume-above-100-percent"]

    if user_settings := _GSettings("org.gnome.desktop.peripherals.mouse"):
        mouse_settings['pointer-acceleration'] = user_settings["accel-profile"]
        mouse_settings['natural-scrolling'] = user_settings["natural-scroll"]
        mouse_settings['speed'] = user_settings["speed"]

    if user_settings := _GSettings("org.gnome.desktop.peripherals.touchpad"):
        touchpad_settings['tap-to-click'] = user_settings["tap-to-click"]
        touchpad_settings['natural-scrolling'] = user_settings["natural-scroll"]
        touchpad_settings['two-finger-scrolling'] = user_settings["two-finger-scrolling-enabled"]
        touchpad_settings['disable-while-typing'] = user_settings["disable-while-typing"]
        touchpad_settings['speed'] = user_settings["speed"]
        if user_settings["send-events"] == 'enabled':
            touchpad_settings['enable'] = True
            touchpad_settings['disable-on-external-mouse'] = False
        elif user_settings["send-events"] == 'disabled':
            touchpad_settings['enable'] = False
            touchpad_settings['disable-on-external-mouse'] = False
        elif user_settings["send-events"] == 'disabled-on-external-mouse':
            touchpad_settings['enable'] = True
            touchpad_settings['disable-on-external-mouse'] = True

    if user_settings := _GSettings("org.gnome.settings-daemon.plugins.power"):
        power_settings['power-button-action'] = user_settings['power-button-action']
        power_settings['auto-power-saver'] = user_settings['power-saver-profile-on-low-battery']
        power_settings['dim-screen'] = user_settings['idle-dim']
        power_settings['suspend-on-ac'] = user_settings['sleep-inactive-ac-type'] == 'suspend'
        power_settings['suspend-on-ac-delay'] = user_settings['sleep-inactive-ac-timeout'] / 60
        power_settings['suspend-on-battery'] = user_settings['sleep-inactive-battery-type'] == 'suspend'
        power_settings['suspend-on-battery-delay'] = user_settings['sleep-inactive-battery-timeout'] / 60

    if user_settings := _GSettings("org.gnome.desktop.session"):
        if user_settings['idle-delay']:
            power_settings['blank-screen'] = True
            power_settings['idle-delay'] = user_settings['idle-delay'] / 60
        else:
            power_settings['blank-screen'] = False

    if user_settings := _GSettings("org.gnome.settings-daemon.plugins.color"):
        night_light_settings['enabled'] = user_settings["night-light-enabled"]
        night_light_settings['schedule-automatic'] = user_settings["night-light-schedule-automatic"]
        night_light_settings['temperature'] = user_settings["night-light-temperature"]

        def hour_minute(decimal_time):
            hour = trunc(decimal_time)
            minute = round((decimal_time % 1) * 60)
            if minute == 60:
                hour += 1
                minute = 0
            return hour, minute

        start_time = user_settings["night-light-schedule-from"]
        start_hour, start_minute = hour_minute(start_time)
        night_light_settings['start-hour'] = start_hour
        night_light_settings['start-minute'] = start_minute

        end_time = user_settings["night-light-schedule-to"]
        end_hour, end_minute = hour_minute(end_time)
        night_light_settings['end-hour'] = end_hour
        night_light_settings['end-minute'] = end_minute

    if user_settings := _GSettings("org.gnome.login-screen"):
        login_screen_settings['enable-welcome-message'] = user_settings["banner-message-enable"]
        login_screen_settings['welcome-message'] = user_settings["banner-message-text"]
        login_screen_settings['logo'] = user_settings["logo"]
        login_screen_settings['enable-logo'] = bool(user_settings["logo"])
        login_screen_settings['disable-restart-buttons'] = user_settings["disable-restart-buttons"]
        login_screen_settings['disable-user-list'] = user_settings["disable-user-list"]
        login_screen_settings['enable-fingerprint-authentication'] = (
            user_settings["enable-fingerprint-authentication"]
        )


def load_from_file(filename=None):
    config_parser = ConfigParser()

    try:
        if filename:
            logger.info(_("Importing from file '{filename}'").format(filename=filename))
            config_parser.read(filename)
        else:
            logger.info(_('Importing from standard input'))
            config_parser.read_file(sys.stdin)
    except ParsingError:
        logger.error(_('Failed to parse import file'))
        raise
    except UnicodeDecodeError:
        logger.error(_('Failed to read import file. Not encoded in UTF-8'))
        raise

    for settings in all_settings:
        section_name = settings.props.schema_id
        if section_name not in config_parser:
            logger.warn(_("Imported file does not have section '{section_name}'"
                          ).format(section_name=section_name))
            continue
        for key in settings:
            if key not in config_parser[section_name]:
                logger.warn(_("Imported file does not have key '{key_name}' in section '{section_name}'"
                              ).format(key_name=key, section_name=section_name))
                continue
            key_type = type(settings[key])
            if key_type == bool:
                settings[key] = config_parser.getboolean(section_name, key)
            else:
                settings[key] = key_type(config_parser[section_name][key])


def save_to_file(filename=None):
    config_parser = ConfigParser()

    for settings in all_settings:
        section_name = settings.props.schema_id
        config_parser[section_name] = {}
        for key in settings:
            config_parser[section_name][key] = str(settings[key])

    if filename:
        logger.info(_("Exporting to file '{filename}'").format(filename=filename))
        try:
            with open(filename, 'w') as outfile:
                config_parser.write(outfile)
        except PermissionError:
            logger.error(_("Cannot write to file '{filename}'. Permission denied"
                           ).format(filename=filename))
            raise
        except IsADirectoryError:
            logger.error(_("Cannot write to file '{filename}'. A directory with "
                            "the same name already exists"
                           ).format(filename=filename))
            raise
    else:
        logger.info(_('Exporting to standard output'))
        config_parser.write(sys.stdout)


def drop_unapplied_changes():
    '''Forget about changes that have not been applied yet'''
    for settings in all_settings:
        settings.revert()


def get_css() -> str:
    '''Get CSS for current settings (to append to theme's 'gnome-shell.css' resource)'''

    css = "\n\n/* 'GDM Settings' App Provided CSS */\n"

    ### Background ###
    background_type = BackgroundType[appearance_settings['background-type']]
    background_image = appearance_settings['background-image']
    adjustment_strat = appearance_settings['bg-adjustment']
    if background_type is BackgroundType.image and background_image:
        css += "\n"
        css += ".login-dialog { background: transparent; }\n"
        css += "#lockDialogGroup {\n"
        css += "  background-image: url('resource:///org/gnome/shell/theme/background');\n"
        if adjustment_strat == 'zoom':
            css += "  background-position: center;\n"
            css += "  background-size: cover;\n"
        else: # adjustment_strat == 'repeat':
            css += "  background-position: 0 0;\n"
            css += "  background-size: contain;\n"
            css += "  background-repeat: repeat;\n"
        css += "}\n"
    elif background_type is BackgroundType.color:
        background_color = appearance_settings['background-color']
        css += "\n"
        css += ".login-dialog { background: transparent; }\n"
        css += "#lockDialogGroup { background-color: " + background_color + "; }\n"

    ### Top Bar ###
    def select_elem(elem:str='') -> str:
        if elem:
            return f"#panel .{elem}, #panel.login-screen .{elem}, #panel.unlock-screen .{elem}"
        else:
            return f"#panel, #panel.login-screen, #panel.unlock-screen"

    disable_arrows = top_bar_settings['disable-arrows']
    disable_rounded_corners = top_bar_settings['disable-rounded-corners']
    change_text_color = top_bar_settings['change-text-color']
    text_color = top_bar_settings['text-color']
    change_background_color = top_bar_settings['change-background-color']
    background_color = top_bar_settings['background-color']

    # Arrows
    if disable_arrows:
        css += "\n"
        css += select_elem("popup-menu-arrow") + " { width: 0px; }\n"
    # Rounded Corners
    if disable_rounded_corners:
        css += "\n"
        css +=  select_elem("panel-corner")
        css +=  " {\n"
        css += f"  -panel-corner-opacity: 0;\n"
        css +=  "}\n"
    # Text Color
    if change_text_color:
        css += "\n"
        css +=  select_elem('panel-button') + ",\n"
        css +=  select_elem('panel-button.clock-display') + " {\n"
        css += f"  color: {text_color};\n"
        css +=  "}\n"
    # Background Color
    if change_background_color:
        css += "\n"
        css +=  select_elem()
        css +=  " {\n"
        css += f"  background-color: {background_color};\n"
        css +=  "}\n"
        if not disable_rounded_corners:
            css += "\n"
            css +=  select_elem("panel-corner")
            css +=  " {\n"
            css += f"  -panel-corner-opacity: 1;\n"
            css += f"  -panel-corner-background-color: {background_color};\n"
            css +=  "}\n"

    # Large Welcome Message
    if (login_screen_settings['enable-welcome-message']
    and login_screen_settings['enlarge-welcome-message']):
        css += "\n"
        css += ".login-dialog-banner {\n"
        css += "  font-size: 1.5em;\n"
        css += "  font-weight: bold;\n"
        css += "}\n"

    return css


def _backup_default_shell_theme ():
    '''back up the default shell theme (if needed)'''

    logger.info(_("Backing up default shell theme …"))

    if gresource.is_unmodified(gresource.ShellGresourceFile):
        _commands.add(f"cp {gresource.ShellGresourceFile} {gresource.DefaultGresourceFile}")

    os.makedirs(env.TEMP_DIR, exist_ok=True)

    gresource.extract_default_theme(f'{env.TEMP_DIR}/default-pure')

    _commands.add(f"rm -rf {gresource.ThemesDir}/default-pure")
    _commands.add(f"mkdir -p {gresource.ThemesDir}")
    _commands.add(f"cp -r {env.TEMP_DIR}/default-pure -t {gresource.ThemesDir}")


def _gresource_apply():
    ''' Apply settings that require modification of 'gnome-shell-theme.gresource' file '''


    # If needed, back up the default shell theme

    pure_theme_not_exists = 'default-pure' not in shell_themes.theme_ids
    shell_gresource_is_stock = gresource.is_unmodified(gresource.ShellGresourceFile)

    if shell_gresource_is_stock or pure_theme_not_exists:
        _backup_default_shell_theme()


    # Apply shell theme settings

    theme_id = appearance_settings['shell-theme']
    theme_path = shell_themes.get_path(theme_id)
    shelldir   = os.path.join(theme_path, 'gnome-shell') if theme_path else None

    background_type = BackgroundType[appearance_settings['background-type']]
    background_image = None
    if background_type is BackgroundType.image:
        background_image = appearance_settings['background-image']

    compiled_file = gresource.compile(shelldir,
          additional_css=get_css(),
        background_image=background_image
    )


    # We need to copy the compiled gresource file instead of moving it because the
    # copy gets correct SELinux context/label where applicable and prevents breakage
    # of GDM in such situations.
    _commands.add(f"install -m644 {compiled_file} {gresource.ShellGresourceFile}")

    if gresource.UbuntuGdmGresourceFile:
        name_of_alternative = os.path.basename(gresource.UbuntuGdmGresourceFile)
        _commands.add('update-alternatives --quiet --install'
                      f' {gresource.UbuntuGdmGresourceFile}'
                      f' {name_of_alternative}'
                      f' {gresource.ShellGresourceFile}'
                      ' 0')

        _commands.add('update-alternatives --quiet --set'
                      f' {name_of_alternative}'
                      f' {gresource.ShellGresourceFile}')


def _dconf_apply():
    ''' Apply settings that are applied through 'dconf' '''

    gdm_conf_dir = "/etc/dconf/db/gdm.d"
    gdm_profile_dir = "/etc/dconf/profile"
    gdm_profile_path = f"{gdm_profile_dir}/gdm"

    temp_profile_path = f"{env.TEMP_DIR}/gdm-profile"
    with open(temp_profile_path, "w+") as temp_profile_file:
        gdm_profile_contents  = "user-db:user\n"
        gdm_profile_contents += "system-db:gdm\n"
        gdm_profile_contents += "file-db:/usr/share/gdm/greeter-dconf-defaults\n"
        temp_profile_file.write(gdm_profile_contents)

    temp_conf_path = f"{env.TEMP_DIR}/95-gdm-settings"
    with open(temp_conf_path, "w+") as temp_conf_file:

        font = font_settings['font']
        hinting = font_settings['hinting']
        icon_theme = appearance_settings['icon-theme']
        time_format = top_bar_settings['time-format']
        cursor_size = pointing_settings['cursor-size']
        cursor_theme = appearance_settings['cursor-theme']
        accent_color = appearance_settings['accent-color']
        color_scheme = "prefer-light" if appearance_settings['light-mode'] else "default"
        show_date = str(top_bar_settings['show-date']).lower()
        show_seconds = str(top_bar_settings['show-seconds']).lower()
        show_weekday = str(top_bar_settings['show-weekday']).lower()
        antialiasing = font_settings['antialiasing']
        scaling_factor = font_settings['scaling-factor']
        show_battery_percentage = str(top_bar_settings['show-battery-percentage']).lower()

        gdm_conf_contents  =  "#-------- Interface ---------\n"
        gdm_conf_contents +=  "[org/gnome/desktop/interface]\n"
        gdm_conf_contents +=  "#----------------------------\n"
        gdm_conf_contents += f"accent-color='{accent_color}'\n"
        gdm_conf_contents += f"color-scheme='{color_scheme}'\n"
        gdm_conf_contents += f"cursor-theme='{cursor_theme}'\n"
        gdm_conf_contents += f"cursor-size={cursor_size}\n"
        gdm_conf_contents += f"icon-theme='{icon_theme}'\n"
        gdm_conf_contents += f"show-battery-percentage={show_battery_percentage}\n"
        gdm_conf_contents += f"clock-show-date={show_date}\n"
        gdm_conf_contents += f"clock-show-seconds={show_seconds}\n"
        gdm_conf_contents += f"clock-show-weekday={show_weekday}\n"
        gdm_conf_contents += f"clock-format='{time_format}'\n"
        gdm_conf_contents += f"font-name='{font}'\n"
        gdm_conf_contents += f"font-antialiasing='{antialiasing}'\n"
        gdm_conf_contents += f"font-hinting='{hinting}'\n"
        gdm_conf_contents += f"text-scaling-factor={scaling_factor}\n"
        gdm_conf_contents +=  "\n"

        accessibility_menu = str(accessibility_settings['always-show-accessibility-menu']).lower()

        gdm_conf_contents +=  "#---- Accessibility ----\n"
        gdm_conf_contents +=  "[org/gnome/desktop/a11y]\n"
        gdm_conf_contents +=  "#------------------------\n"
        gdm_conf_contents += f"always-show-universal-access-status={accessibility_menu}\n"
        gdm_conf_contents +=  "\n"

        sound_theme = sound_settings['theme']
        event_sounds = str(sound_settings['event-sounds']).lower()
        feedback_sounds = str(sound_settings['feedback-sounds']).lower()
        over_amplification = str(sound_settings['over-amplification']).lower()

        gdm_conf_contents +=  "#-------- Sound ---------\n"
        gdm_conf_contents +=  "[org/gnome/desktop/sound]\n"
        gdm_conf_contents +=  "#------------------------\n"
        gdm_conf_contents += f"theme-name='{sound_theme}'\n"
        gdm_conf_contents += f"event-sounds={event_sounds}\n"
        gdm_conf_contents += f"input-feedback-sounds={feedback_sounds}\n"
        gdm_conf_contents += f"allow-volume-above-100-percent={over_amplification}\n"
        gdm_conf_contents +=  "\n"

        pointer_acceleration = mouse_settings['pointer-acceleration']
        natural_scrolling = str(mouse_settings['natural-scrolling']).lower()
        mouse_speed = mouse_settings['speed']

        gdm_conf_contents +=  "#-------------- Mouse ---------------\n"
        gdm_conf_contents +=  "[org/gnome/desktop/peripherals/mouse]\n"
        gdm_conf_contents +=  "#------------------------------------\n"
        gdm_conf_contents += f"accel-profile='{pointer_acceleration}'\n"
        gdm_conf_contents += f"natural-scroll={natural_scrolling}\n"
        gdm_conf_contents += f"speed={mouse_speed}\n"
        gdm_conf_contents +=  "\n"

        touchpad_speed = touchpad_settings['speed']
        tap_to_click = str(touchpad_settings['tap-to-click']).lower()
        natural_scrolling = str(touchpad_settings['natural-scrolling']).lower()
        two_finger_scrolling = str(touchpad_settings['two-finger-scrolling']).lower()
        disable_while_typing = str(touchpad_settings['disable-while-typing']).lower()
        enable_touchpad = touchpad_settings['enable']
        disable_on_external_mouse = touchpad_settings['disable-on-external-mouse']

        gdm_conf_contents +=  "#-------------- Touchpad ---------------\n"
        gdm_conf_contents +=  "[org/gnome/desktop/peripherals/touchpad]\n"
        gdm_conf_contents +=  "#---------------------------------------\n"
        gdm_conf_contents += f"speed={touchpad_speed}\n"
        gdm_conf_contents += f"tap-to-click={tap_to_click}\n"
        gdm_conf_contents += f"natural-scroll={natural_scrolling}\n"
        gdm_conf_contents += f"two-finger-scrolling-enabled={two_finger_scrolling}\n"
        gdm_conf_contents += f"disable-while-typing={disable_while_typing}\n"
        if not enable_touchpad:
            gdm_conf_contents += f"send-events='disabled'\n"
        elif disable_on_external_mouse:
             gdm_conf_contents += f"send-events='disabled-on-external-mouse'\n"
        else:
             gdm_conf_contents += f"send-events='enabled'\n"


        gdm_conf_contents +=  "\n"

        power_button_action = power_settings['power-button-action']
        auto_power_saver = str(power_settings['auto-power-saver']).lower()
        dim_screen = str(power_settings['dim-screen']).lower()
        idle_delay = int(power_settings['idle-delay'] * 60) if power_settings['blank-screen'] else 0
        sleep_type_on_ac = 'suspend' if power_settings['suspend-on-ac'] else 'nothing'
        suspend_on_ac_delay = int(power_settings['suspend-on-ac-delay'] * 60)
        sleep_type_on_battery = 'suspend' if power_settings['suspend-on-battery'] else 'nothing'
        suspend_on_battery_delay = int(power_settings['suspend-on-battery-delay'] * 60)

        gdm_conf_contents +=  "#---------------- Power -----------------\n"
        gdm_conf_contents +=  "[org/gnome/settings-daemon/plugins/power]\n"
        gdm_conf_contents +=  "#----------------------------------------\n"
        gdm_conf_contents += f"power-button-action='{power_button_action}'\n"
        gdm_conf_contents += f"power-saver-profile-on-low-battery={auto_power_saver}\n"
        gdm_conf_contents += f"dim-screen={dim_screen}\n"
        gdm_conf_contents += f"sleep-inactive-ac-type='{sleep_type_on_ac}'\n"
        gdm_conf_contents += f"sleep-inactive-ac-timeout={suspend_on_ac_delay}\n"
        gdm_conf_contents += f"sleep-inactive-battery-type='{sleep_type_on_battery}'\n"
        gdm_conf_contents += f"sleep-inactive-battery-timeout={suspend_on_battery_delay}\n"
        gdm_conf_contents +=  "\n"
        gdm_conf_contents +=  "#--------------- Session ----------------\n"
        gdm_conf_contents +=  "[org/gnome/desktop/session]\n"
        gdm_conf_contents +=  "#----------------------------------------\n"
        gdm_conf_contents += f"idle-delay={idle_delay}\n"
        gdm_conf_contents +=  "\n"

        enabled = str(night_light_settings['enabled']).lower()
        temperature = round(night_light_settings['temperature'])
        schedule_automatic = str(night_light_settings['schedule-automatic']).lower()
        schedule_from  = night_light_settings['start-hour']
        schedule_from += night_light_settings['start-minute'] / 60
        schedule_to  = night_light_settings['end-hour']
        schedule_to += night_light_settings['end-minute'] / 60

        gdm_conf_contents +=  "#------------- Night Light --------------\n"
        gdm_conf_contents +=  "[org/gnome/settings-daemon/plugins/color]\n"
        gdm_conf_contents +=  "#----------------------------------------\n"
        gdm_conf_contents += f"night-light-enabled={enabled}\n"
        gdm_conf_contents += f"night-light-temperature=uint32 {temperature}\n"
        gdm_conf_contents += f"night-light-schedule-automatic={schedule_automatic}\n"
        gdm_conf_contents += f"night-light-schedule-from={schedule_from}\n"
        gdm_conf_contents += f"night-light-schedule-to={schedule_to}\n"
        gdm_conf_contents +=  "\n"

        enable_logo = login_screen_settings['enable-logo']
        logo = '/etc/gdm-logo' if enable_logo else ''
        logo_file = login_screen_settings['logo'].removeprefix(env.HOST_ROOT)
        enable_welcome_message = str(login_screen_settings['enable-welcome-message']).lower()
        welcome_message = login_screen_settings['welcome-message'].replace("'", r"\'")
        disable_restart_buttons = str(login_screen_settings['disable-restart-buttons']).lower()
        disable_user_list = str(login_screen_settings['disable-user-list']).lower()
        enable_fingerprint_authentication = (
            str(login_screen_settings['enable-fingerprint-authentication']).lower()
        )

        gdm_conf_contents +=  "#----- Login Screen ----\n"
        gdm_conf_contents +=  "[org/gnome/login-screen]\n"
        gdm_conf_contents +=  "#-----------------------\n"
        gdm_conf_contents += f"logo='{logo}'\n"
        gdm_conf_contents += f"banner-message-enable={enable_welcome_message}\n"
        gdm_conf_contents += f"banner-message-text='{welcome_message}'\n"
        gdm_conf_contents += f"disable-restart-buttons={disable_restart_buttons}\n"
        gdm_conf_contents += f"disable-user-list={disable_user_list}\n"
        gdm_conf_contents += (
            f"enable-fingerprint-authentication={enable_fingerprint_authentication}\n"
        )
        
        temp_conf_file.write(gdm_conf_contents)

    if enable_logo and logo_file:
        if not os.path.exists(logo_file):
            raise LogoImageNotFoundError(2, 'No such file', logo_file)

        logo_temp = os.path.join(env.TEMP_DIR, 'logo.temp')
        shutil.copy(logo_file, logo_temp)
        _commands.add(f"install -m644 '{logo_temp}' -T '{logo}'")

    overriding_files = get_overriding_files()
    if overriding_files:
        _commands.add('rm', *overriding_files)

    _commands.add(f"install -Dm644 '{temp_conf_path}' -t '{gdm_conf_dir}'")
    _commands.add(f"install -Dm644 '{temp_profile_path}' -T '{gdm_profile_path}'")
    _commands.add("dconf update")


def apply() -> bool:
    ''' Apply all settings '''

    _gresource_apply()
    _dconf_apply()

    result = _commands.run()

    if result:
        # When we change GDM shell theme it becomes the default theme but for the users
        # who didn't want to change shell theme for their session, we need to set it to a
        # pure/original version of the default shell theme
        # Note: We don't want to change user's shell theme if user set it explicitly to
        # default in order to match their GDM theme
        if user_settings := _GSettings('org.gnome.shell.extensions.user-theme'):
            if (user_settings['name'] == ''
            and main_settings["never-applied"]):
                user_settings['name'] = 'default-pure'

        main_settings["never-applied"] = False

        '''Save settings to GSettings of this app'''
        for settings in all_settings:
            settings.apply()

    return result


def apply_user_display_settings() -> bool:
    ''' Apply user's current display settings '''

    user_monitors_xml = os.path.join(env.XDG_CONFIG_HOME, 'monitors.xml')

    if not os.path.isfile(user_monitors_xml):
        raise FileNotFoundError(2, 'No such file or directory', user_monitors_xml)

    with open(user_monitors_xml) as monitors_xml:
        if monitors_xml.read() == '':
            raise FileNotFoundError(2, 'File is empty', user_monitors_xml)

    temp_monitors_xml = os.path.join(env.TEMP_DIR, 'monitors.xml')
    shutil.copyfile(user_monitors_xml, temp_monitors_xml)
    os.chmod(temp_monitors_xml, 0o644)

    experimental_features = []
    if user_settings := _GSettings('org.gnome.mutter'):
        experimental_features = user_settings['experimental-features']


    _commands.add(['machinectl', 'shell', f'{gresource.GdmUsername}@', '/usr/bin/env',
                     'gsettings', 'set', 'org.gnome.mutter', 'experimental-features',
                     f'"{experimental_features}"',
                     '&>/dev/null',
                   ])

    _commands.add(['install', '-Dm644',
                     '-o', gresource.GdmUsername,
                     temp_monitors_xml,
                     f'~{gresource.GdmUsername}/.config/monitors.xml',
                   ])

    print(_commands)
    return _commands.run()


def reset() -> bool:
    if gresource.UbuntuGdmGresourceFile:
        _commands.add(['update-alternatives',  '--quiet',  '--remove',
                          os.path.basename(gresource.UbuntuGdmGresourceFile),
                          gresource.ShellGresourceFile,
                       ])

    if os.path.exists(gresource.DefaultGresourceFile):
        _commands.add(['mv', '-f', gresource.DefaultGresourceFile, gresource.ShellGresourceFile])

    _commands.add("rm -f /etc/dconf/profile/gdm")
    _commands.add("rm -f /etc/dconf/db/gdm.d/95-gdm-settings")
    _commands.add("dconf update")

    if _commands.run():

        for settings in all_settings:
            for key in settings.props.settings_schema.list_keys():
                settings.reset(key)
            settings.apply()

        return True

    return False


def get_overriding_files():
    gdm_conf_dir = "/etc/dconf/db/gdm.d"
    our_config = os.path.join(gdm_conf_dir, '95-gdm-settings')
    overriding_files = []

    if os.path.isdir (env.HOST_ROOT + gdm_conf_dir):
        files = set()
        for dirpath, dirnames, filenames in os.walk(env.HOST_ROOT + gdm_conf_dir):
            for filename in filenames:
                files.add(os.path.join(dirpath.removeprefix(env.HOST_ROOT), filename))
        files.add(our_config)

        sorted_files = sorted (files)
        index_of_next_file = sorted_files.index(our_config) + 1
        overriding_files = sorted_files[index_of_next_file:]

    return overriding_files



from gi.repository import Gtk

from gdms.utils import GProperty
from gdms.settings import pointing_settings, mouse_settings, touchpad_settings

from .common import PageContent


class PointingPageContent (PageContent):
    __gtype_name__ = 'PointingPageContent'

    def __init__ (self, window, **props):
        super().__init__(**props)

        self.window = window

        self.builder = Gtk.Builder.new_from_resource('/app/ui/pointing-page.ui')

        self.set_child(self.builder.get_object('content_box'))

        # General
        self.cursor_size_selector = self.builder.get_object('cursor_size_selector')
        # Mouse
        self.m_acceleration_comborow = self.builder.get_object('m_acceleration_comborow')
        self.m_natural_scrolling_row = self.builder.get_object('m_natural_scrolling_row')
        self.m_speed_scale = self.builder.get_object('m_speed_scale')
        # Touchpad
        self.t_tap_to_click_row = self.builder.get_object('t_tap_to_click_row')
        self.t_natural_scrolling_row = self.builder.get_object('t_natural_scrolling_row')
        self.t_two_finger_scrolling_row = self.builder.get_object('t_two_finger_scrolling_row')
        self.t_disable_while_typing_row = self.builder.get_object('t_disable_while_typing_row')
        self.t_speed_scale = self.builder.get_object('t_speed_scale')
        self.t_speed_scalerow= self.builder.get_object('t_speed_scalerow')
        self.t_enable_switch = self.builder.get_object('t_enable_switch')
        self.t_disable_on_external_mouse_row = self.builder.get_object('t_disable_on_external_mouse_row')

        self.t_enable_switch.connect("notify::active", self.on_t_enable_switch_notify_active);

        # Following properties are ignored when set in .ui files.
        # So, they need to be changed here.
        self.m_speed_scale.set_range(-1, 1)
        self.t_speed_scale.set_range(-1, 1)

        self.bind_to_gsettings()
        self.on_t_enable_switch_notify_active(self.t_enable_switch, None)


    def on_t_enable_switch_notify_active(self, switch: Gtk.Switch, data):
        if switch.props.active:
            self.t_disable_on_external_mouse_row.set_sensitive(True)
            self.t_tap_to_click_row.set_sensitive(True)
            self.t_natural_scrolling_row.set_sensitive(True)
            self.t_two_finger_scrolling_row.set_sensitive(True)
            self.t_disable_while_typing_row.set_sensitive(True)
            self.t_speed_scalerow.set_sensitive(True)
        else:
            self.t_disable_on_external_mouse_row.set_sensitive(False)
            self.t_tap_to_click_row.set_sensitive(False)
            self.t_natural_scrolling_row.set_sensitive(False)
            self.t_two_finger_scrolling_row.set_sensitive(False)
            self.t_disable_while_typing_row.set_sensitive(False)
            self.t_speed_scalerow.set_sensitive(False)

    def bind_to_gsettings (self):
        # General
        pointing_settings.bind('cursor-size', self.cursor_size_selector, 'selected-size')
        # Mouse
        mouse_settings.bind_via_list('pointer-acceleration', self.m_acceleration_comborow, 'selected',
                                     ['default', 'flat', 'adaptive'])
        mouse_settings.bind('natural-scrolling', self.m_natural_scrolling_row, 'active')
        mouse_settings.bind('speed', self.m_speed_scale.props.adjustment, 'value')
        # Touchpad
        touchpad_settings.bind('tap-to-click', self.t_tap_to_click_row, 'active')
        touchpad_settings.bind('natural-scrolling', self.t_natural_scrolling_row, 'active')
        touchpad_settings.bind('two-finger-scrolling', self.t_two_finger_scrolling_row, 'active')
        touchpad_settings.bind('disable-while-typing', self.t_disable_while_typing_row, 'active')
        touchpad_settings.bind('speed', self.t_speed_scale.props.adjustment, 'value')
        touchpad_settings.bind('enable', self.t_enable_switch, 'active')
        touchpad_settings.bind('disable-on-external-mouse', self.t_disable_on_external_mouse_row, 'active')

@Gtk.Template(resource_path='/app/ui/pointing-page/cursor-size-button.ui')
class CursorSizeButton (Gtk.ToggleButton):
    __gtype_name__ = 'CursorSizeButton'

    cursor_size = GProperty(int, default=24, construct_only=True)
    size_name = GProperty(str)


@Gtk.Template(resource_path='/app/ui/pointing-page/cursor-size-selector.ui')
class CursorSizeSelector (Gtk.ListBoxRow, Gtk.Buildable):
    __gtype_name__ = 'CursorSizeSelector'
    box = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

        self._item_dict = {}
        self._first_item = None
        self._selected_size = 0
        self._selected_name = ''

    @GProperty(int)
    def selected_size(self):
        return self._selected_size

    @selected_size.setter
    def selected_size(self, value):
        if self._selected_size == value:
            return

        if value not in self._item_dict:
            return

        self._selected_size = value
        btn = self._item_dict[value]
        btn.set_active(True)

        self._selected_name = btn.size_name
        self.notify('selected-name')

    @GProperty(str, writable=False)
    def selected_name(self):
        return self._selected_name

    def do_add_child(self, builder, child, _type):
        if not isinstance(child, CursorSizeButton):
            raise TypeError('Child is a ' + type(child) + '. Should be a CursorSizeButton instead')

        cursor_size = child.cursor_size
        if cursor_size in self._item_dict:
            raise ValueError('A CursorSizeButton with cursor-size of ' + cursor_size
                             + ' has already been added')

        self.box.append(child)
        child.connect('toggled', self.selection_changed_cb)
        self._item_dict[cursor_size] = child

        if self._first_item:
            child.set_group(self._first_item)
        else:
            self._first_item = child
            child.set_active(True)

    def selection_changed_cb(self, child):
        if not child.get_active() or child.cursor_size == self.selected_size:
            return

        self.selected_size = child.cursor_size



from .appearance   import AppearancePageContent
from .fonts        import FontsPageContent
from .top_bar      import TopBarPageContent
from .sound        import SoundPageContent
from .pointing     import PointingPageContent
from .display      import DisplayPageContent
from .login_screen import LoginScreenPageContent
from .tools        import ToolsPageContent
from .power        import PowerPageContent



''' Utilities (functions) for GResource files of GNOME Shell themes'''

import os
import shutil
import subprocess

from gdms import env
from gdms import utils


ThemesDir              = os.path.join(env.HOST_DATA_DIRS[0], 'themes')
CustomThemeIdentity    = 'custom-theme'
GdmUsername            = 'gdm'
ShellGresourceFile     = None
DefaultGresourceFile   = None
UbuntuGdmGresourceFile = None

for data_dir in env.HOST_DATA_DIRS:
    file = os.path.join (data_dir,  'gnome-shell', 'gnome-shell-theme.gresource')
    if os.path.isfile (env.HOST_ROOT + file):
        ShellGresourceFile   = file
        DefaultGresourceFile = ShellGresourceFile + ".default"
        break

if 'ubuntu' in env.OS_IDs:
    if os.path.exists(env.HOST_ROOT+'/usr/share/gnome-shell/gdm-theme.gresource'):
        UbuntuGdmGresourceFile = '/usr/share/gnome-shell/gdm-theme.gresource'
    else:
        UbuntuGdmGresourceFile = '/usr/share/gnome-shell/gdm3-theme.gresource'

with open(env.HOST_ROOT+'/etc/passwd') as passwd_db:
    for line in passwd_db:
        if line.startswith('Debian-gdm'):
            GdmUsername = 'Debian-gdm'
            break


def is_unmodified(gresourceFile:str):
    """checks if the provided file is a GResource file of the default theme"""

    if env.HOST_ROOT and not gresourceFile.startswith(env.HOST_ROOT):
        gresourceFile = env.HOST_ROOT + gresourceFile

    gresource_contains = lambda x: utils.get_stdout(["gresource", "list", gresourceFile, "/org/gnome/shell/theme/" + x])

    if (os.path.exists(gresourceFile) and
        (gresource_contains("gnome-shell.css") or gresource_contains("gnome-shell-dark.css")) and
        not gresource_contains(CustomThemeIdentity)
        ):
            return True
    return False

def get_default() -> str:
    """get full path to the unmodified GResource file of the default theme (if the file exists)"""

    for file in ShellGresourceFile, DefaultGresourceFile:
       if is_unmodified(file):
           return file

def extract_default_theme(destination:str, /):
    '''Extract default GNOME Shell theme'''

    if os.path.exists(destination):
        shutil.rmtree(destination)

    gresource_file = get_default()

    if not gresource_file:
        raise FileNotFoundError('No unmodified GResource file of the default shell theme was found')

    destination_shell_dir = os.path.join(destination, 'gnome-shell')
    resource_list = utils.get_stdout(["gresource", "list", env.HOST_ROOT + gresource_file]).splitlines()

    for resource in resource_list:
        filename = resource.removeprefix("/org/gnome/shell/theme/")
        filepath = os.path.join(destination_shell_dir, filename)
        content  = utils.get_stdout(["gresource", "extract", env.HOST_ROOT + gresource_file, resource],
                                  decode=False)

        os.makedirs(os.path.dirname(filepath), exist_ok=True)

        with open(filepath, "wb") as opened_file:
            opened_file.write(content)

class BackgroundImageNotFoundError (FileNotFoundError): pass

def append_to_file(path:str, contents:str):
    with open(path, 'a') as f:
        f.write(contents)

def append_if_exists(path:str, contents:str):
    if os.path.isfile(path):
        append_to_file(path, contents)

def compile(shellDir:str, additional_css:str, background_image:str=''):
    """Compile a theme into a GResource file for its use as a GDM theme"""

    temp_gresource_file = os.path.join(env.TEMP_DIR, 'gnome-shell-theme.gresource')
    temp_gresource_xml = f'{temp_gresource_file}.xml'
    temp_theme_dir = os.path.join(env.TEMP_DIR, 'temp-theme')
    temp_shell_dir = os.path.join(temp_theme_dir, 'gnome-shell')

    if os.path.exists(temp_theme_dir):
        shutil.rmtree(temp_theme_dir)

    if os.path.exists(temp_gresource_file):
        os.remove(temp_gresource_file)

    extract_default_theme(temp_theme_dir)

    if shellDir:
        css_files = {}
        for filename in os.listdir(temp_shell_dir):
            if not filename.endswith('.css'):
                continue
            with open(os.path.join(temp_shell_dir, filename)) as file_io:
                css_files[filename] = file_io.read()

        shutil.copytree(shellDir, temp_shell_dir, dirs_exist_ok=True)

        shelldir_has_file = lambda x: os.path.isfile(os.path.join(shellDir, x))

        for filename, contents_default in css_files.items():
            if not shelldir_has_file(filename):
                continue
            with open(os.path.join(shellDir, filename)) as file_io:
                contents_theme = file_io.read()
            with open(os.path.join(temp_shell_dir, filename), 'w') as file_io:
                contents = contents_default + '\n\n' + contents_theme
                file_io.write(contents)

        if shelldir_has_file("gnome-shell.css"):
            if "gnome-shell-dark.css" in css_files and not shelldir_has_file("gnome-shell-dark.css"):
                with open(os.path.join(shellDir, "gnome-shell.css")) as file_io:
                    contents_theme = file_io.read()
                with open(os.path.join(temp_shell_dir, "gnome-shell-dark.css"), 'w') as file_io:
                    contents = css_files["gnome-shell-dark.css"] + '\n\n' + contents_theme
                    file_io.write(contents)

            if "gnome-shell-light.css" in css_files and not shelldir_has_file("gnome-shell-light.css"):
                with open(os.path.join(shellDir, "gnome-shell.css")) as file_io:
                    contents_theme = file_io.read()
                with open(os.path.join(temp_shell_dir, "gnome-shell-light.css"), 'w') as file_io:
                    contents = css_files["gnome-shell-light.css"] + '\n\n' + contents_theme
                    file_io.write(contents)

            if "gnome-shell.css" not in css_files:
                shutil.copy(f"{temp_shell_dir}/gnome-shell-dark.css", f"{temp_shell_dir}/gnome-shell.css")

        elif shelldir_has_file("gnome-shell-dark.css") and "gnome-shell.css" in css_files:
            with open(os.path.join(shellDir, "gnome-shell-dark.css")) as file_io:
                contents_theme = file_io.read()
            with open(os.path.join(temp_shell_dir, "gnome-shell.css"), 'w') as file_io:
                contents = css_files["gnome-shell.css"] + '\n\n' + contents_theme
                file_io.write(contents)

    # Inject custom-theme identity
    open(os.path.join(temp_shell_dir, CustomThemeIdentity), 'w').close()

    if background_image:
        if os.path.isfile(background_image):
            shutil.copy(background_image, os.path.join(temp_shell_dir, 'background'))
        else:
            raise BackgroundImageNotFoundError(2, 'No such file', background_image)

    append_if_exists(f"{temp_shell_dir}/gnome-shell.css", additional_css)
    append_if_exists(f"{temp_shell_dir}/gnome-shell-dark.css", additional_css)
    append_if_exists(f"{temp_shell_dir}/gnome-shell-light.css", additional_css)
    append_if_exists(f"{temp_shell_dir}/gnome-shell-high-contrast.css", additional_css)

    if not os.path.isfile(os.path.join(temp_shell_dir, 'gnome-shell.css')):
        shutil.copy(f"{temp_shell_dir}/gnome-shell-dark.css", f"{temp_shell_dir}/gnome-shell.css")

    shutil.copy(f"{temp_shell_dir}/gnome-shell.css", f"{temp_shell_dir}/gdm.css")
    shutil.copy(f"{temp_shell_dir}/gnome-shell.css", f"{temp_shell_dir}/gdm3.css")

    with open(temp_gresource_xml, 'w') as GresourceXml:
        print('<?xml version="1.0" encoding="UTF-8"?>',
              '<gresources>',
              ' <gresource prefix="/org/gnome/shell/theme">',
            *('  <file>'+file+'</file>' for file in utils.list_files(temp_shell_dir)),
              ' </gresource>',
              '</gresources>',

              sep='\n',
              file=GresourceXml,
             )

    # Compile Theme
    subprocess.run(['glib-compile-resources',
         '--sourcedir', temp_shell_dir,
         '--target', temp_gresource_file,
         temp_gresource_xml,
       ])

    # Return path to the generated GResource file
    return  temp_gresource_file



'''A self-contained collection of general purpose and reusable utilities
(functions & classes)

'self-contained' means that this module does not depend on anything from
the gdm_settings package'''

import os
import pathlib
import subprocess

from enum import EnumType, Enum
from typing import Any, Optional, TypeVar, cast
from collections.abc import Callable, Iterator, Sequence

from gi.repository import Adw
from gi.repository import Gtk
from gi.repository import Pango
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GObject


def get_stdout(command: str | list [str],
               /, *,
               decode: bool = True,
               ) -> str | bytes:
    '''get standard output of a command'''

    if isinstance(command, str):
        command = command.split()

    stdout = subprocess.run(command, stdout=subprocess.PIPE).stdout

    if decode:
        stdout = stdout.decode()

    return stdout


def list_files(dir_path: str,
               base: Optional[str] = None,
               *,
               recursive: bool = True,
               ) -> Iterator[str]:
    """list files (only) inside a directory"""

    if base is None:
        base = dir_path

    for child in os.listdir(dir_path):
        child_path = os.path.join(dir_path, child)

        if not os.path.isdir(child_path):
            yield os.path.relpath(child_path, base)
        elif recursive:
            yield from list_files(child_path, base)


class Path(pathlib.PosixPath):
    def __add__(self, other: pathlib.Path | str):
        if not isinstance(other, (pathlib.Path, str)):
            return NotImplemented

        return Path(str(self) + str(other))

    def __iadd__(self, other: pathlib.Path | str):
        return self + other

    def __radd__(self, other: str):
        if not isinstance(other, str):
            return NotImplemented

        return other + str(self)
    
    def with_prefix(self, prefix: pathlib.Path | str):
        if not isinstance(prefix, (str, pathlib.Path)):
            raise TypeError("prefix must be str or pathlib.Path, not " + repr(type(prefix)))
        
        if not self.is_absolute():
            raise ValueError("adding prefix to non absolute paths is not allowed")

        if str(self).startswith(str(prefix)):
            return self
        
        return Path(prefix) + self
    
    remove = pathlib.PosixPath.unlink

    def append_bytes(self, data: bytes):
        with self.open('ab') as io:
            io.write(data)

    def append_text(self, data: str):
        with self.open('a') as io:
            io.write(data)

    def _list_files(self, base, recursive=True):
        for child in self.iterdir():
            if not child.is_dir():
                yield child.relative_to(base)
            elif recursive:
                yield from child._list_files(base)

    def list_files(self, *, recursive=True):
        """list only the files inside a directory"""

        return self._list_files(base=self, recursive=recursive)


def GProperty(type: type,
              default: Optional[object] = None,
              *args,
              readable: bool = True,
              writable: bool = True,
              construct: bool = False,
              construct_only: bool = False,
              additional_flags: GObject.ParamFlags = GObject.ParamFlags(0),
              **kwargs,
              ) -> GObject.Property:
    '''A wrapper around GObject.Property decorator

    Provides shorter syntax for creating GObject properties.
    '''

    flags = additional_flags
    if readable:
        flags |= GObject.ParamFlags.READABLE
    if writable:
        flags |= GObject.ParamFlags.WRITABLE
    if construct:
        flags |= GObject.ParamFlags.CONSTRUCT
    if construct_only:
        flags |= GObject.ParamFlags.CONSTRUCT_ONLY

    return GObject.Property(*args, type=type, default=default, flags=flags, **kwargs)


Callback = Callable[[], Any]

class InvalidGioTaskError (Exception): pass
class AlreadyRunningError (Exception): pass


class BackgroundTask (GObject.Object):
    __gtype_name__ = 'BackgroundTask'

    def __init__ (self, function: Callback, finish_callback: Callback):
        super().__init__()

        self.function = function
        self.finish_callback = finish_callback
        self._current = None

    def start(self) -> None:
        if self._current:
            AlreadyRunningError('Task is already running')

        finish_callback = lambda self, task, nothing: self.finish_callback()

        task = Gio.Task.new(self, None, finish_callback, None)
        task.run_in_thread(self._thread_cb)

        self._current = task

    @staticmethod
    def _thread_cb (task: Gio.Task, self: GObject.Object, task_data: Any, cancellable: Gio.Cancellable | None):
        try:
            assert isinstance(self, BackgroundTask)
            retval = self.function()
            task.return_value(retval)
        except Exception as e:
            task.return_value(e)

    def finish (self):
        task = self._current
        self._current = None

        if task is None:
            return

        if not Gio.Task.is_valid(task, self):
            raise InvalidGioTaskError()

        _, value = task.propagate_value()

        if isinstance(value, Exception):
            raise value

        return value


class _MappingFuncs:
    '''A collection of functions for 'GSettings' class that convert values
    from Settings keys to GObject properties and vice versa.'''

    class FontDescription:
        @staticmethod
        def from_string(string: str, user_data=None) -> Pango.FontDescription:
            return Pango.FontDescription.from_string(string)

        @staticmethod
        def to_string(font_desc: Pango.FontDescription, user_data=None) -> str:
            return font_desc.to_string()

    class RGBA:
        @staticmethod
        def from_string(string: str, user_data=None) -> Gdk.RGBA:
            rgba = Gdk.RGBA()
            rgba.parse(string)
            return rgba

        @staticmethod
        def to_string(rgba: Gdk.RGBA, user_data=None) -> str:
            return rgba.to_string()

    class ComboRow:
        @staticmethod
        def string_to_selected(string: str, comborow: Adw.ComboRow) -> int:
            string_list = cast(Gtk.StringList, comborow.get_model())
            for position, string_object in enumerate(string_list):
                if string == cast(Gtk.StringObject, string_object).get_string():
                    return position
            return 0

        @staticmethod
        def selected_to_string(position: int, comborow: Adw.ComboRow) -> str:
            selected_item = cast(Gtk.StringObject, comborow.get_selected_item())
            return selected_item.get_string()

    class Enum:
        @staticmethod
        def name_to_value(name: str, enum: EnumType) -> Any:
            return cast(Enum, enum[name]).value
        @staticmethod
        def value_to_name(value: Any, enum: EnumType) -> str:
            return cast(Enum, enum(value)).name

    class List:
        @staticmethod
        def value_to_index_non_strict(value: Any, lyst: Sequence) -> int:
            try:
                return lyst.index(value)
            except ValueError:
                return 0

        @staticmethod
        def value_to_index_strict (value: Any, lyst: Sequence) -> int:
            return lyst.index(value)

        @staticmethod
        def index_to_value (index: int, lyst: Sequence) -> Any:
            return lyst[index]


class GSettings(Gio.Settings):
    '''A wrapper around Gio.Settings for easy usage in Python'''

    __gtype_name__ = "_Settings"    # "GSettings" name is not available
    _default_flag = Gio.SettingsBindFlags.DEFAULT

    def __init__(self, schema_id: str | None = None, **props):
        if (schema_id is not None):
            super().__init__(schema_id=schema_id, **props)
        else:
            super().__init__(**props)

    @classmethod
    def new(cls, schema_id: str):
        return cls(schema_id)

    @classmethod
    def new_delayed(cls, schema_id: str):
        obj = cls(schema_id)
        obj.delay()
        return obj

    @classmethod
    def new_full(cls, schema: Gio.SettingsSchema, backend: Gio.SettingsBackend, path: str):
        return cls(settings_schema=schema, backend=backend, path=path)

    @classmethod
    def new_with_backend(cls, schema_id: str, backend: Gio.SettingsBackend):
        return cls(schema_id=schema_id, backend=backend)

    @classmethod
    def new_with_backend_and_path(cls, schema_id: str, backend: Gio.SettingsBackend, path: str):
        return cls(schema_id=schema_id, backend=backend, path=path)

    @classmethod
    def new_with_path(cls, schema_id: str, path: str):
        return cls(schema_id=schema_id, path=path)

    def bind (self, key: str, obj: GObject.Object, prop: str,
              flags: Gio.SettingsBindFlags = _default_flag) -> None:
        super().bind(key, obj, prop, flags)

    def bind_to_fontbutton(self, key: str, fontbutton: Gtk.FontDialogButton,
                            flags: Gio.SettingsBindFlags = _default_flag,
                            ) -> None:
        self.bind_with_mapping(key, fontbutton, 'font-desc', flags,
                               _MappingFuncs.FontDescription.from_string,
                               _MappingFuncs.FontDescription.to_string)

    def bind_to_colorbutton(self, key: str, colorbutton: Gtk.ColorDialogButton,
                            flags: Gio.SettingsBindFlags = _default_flag,
                            ) -> None:
        self.bind_with_mapping(key, colorbutton, 'rgba', flags,
                               _MappingFuncs.RGBA.from_string,
                               _MappingFuncs.RGBA.to_string)

    def bind_to_comborow(self, key: str, comborow: Adw.ComboRow,
                         flags: Gio.SettingsBindFlags = _default_flag,
                         ) -> None:
        self.bind_with_mapping(key, comborow, 'selected', flags,
                               _MappingFuncs.ComboRow.string_to_selected,
                               _MappingFuncs.ComboRow.selected_to_string,
                               comborow)

    def bind_via_enum(self, key: str, obj: GObject.Object, prop: str, enum: EnumType,
                      flags: Gio.SettingsBindFlags = _default_flag,
                      ) -> None:
        self.bind_with_mapping(key, obj, prop, flags,
                               _MappingFuncs.Enum.name_to_value,
                               _MappingFuncs.Enum.value_to_name,
                               enum)

    def bind_via_list(self, key: str, obj: GObject.Object, prop: str, lyst: Sequence,
                      flags: Gio.SettingsBindFlags = _default_flag,
                      strict: bool = True) -> None:

        value_to_index = _MappingFuncs.List.value_to_index_strict
        index_to_value = _MappingFuncs.List.index_to_value

        if not strict:
            value_to_index = _MappingFuncs.List.value_to_index_non_strict

        self.bind_with_mapping(key, obj, prop, flags,
                               value_to_index,
                               index_to_value,
                               lyst)

    UserData = TypeVar("UserData")
    BindingKey = TypeVar("BindingKey")
    BindingProp = TypeVar("BindingProp")

    def bind_with_mapping(self, key: str, obj: GObject.Object, prop: str,
                          flags: Gio.SettingsBindFlags = _default_flag,
                          key_to_prop: Callable[[BindingKey, UserData], BindingProp] | None = None,
                          prop_to_key: Callable[[BindingProp, UserData], BindingKey] | None = None,
                          user_data: UserData = None,
                          ) -> None:
        '''
        Recreate g_settings_bind_with_mapping since it does not exist in python bindings for Gio.Settings

        Taken directly from
        https://gitlab.gnome.org/GNOME/pygobject/uploads/bdd6ef6ee51a50f82d75673e43c1395b
        /0001-Implement-g_settings_bind_with_mapping-and-handle-ra.patch
        found in issue https://gitlab.gnome.org/GNOME/pygobject/-/issues/98
        '''

        self._ignore_key_changed = False
        self._ignore_prop_changed = False

        def key_changed(self, key: str) -> None:
            if self._ignore_key_changed:
                return
            self._ignore_prop_changed = True
            assert key_to_prop is not None
            obj.set_property(prop, key_to_prop(self[key], user_data))
            self._ignore_prop_changed = False

        def prop_changed(obj: GObject.Object, param: GObject.ParamSpec) -> None:
            if self._ignore_prop_changed:
                return
            self._ignore_key_changed = True
            assert prop_to_key is not None
            cast(dict, self)[key] = prop_to_key(obj.get_property(prop), user_data)
            self._ignore_key_changed = False

        if not (key_to_prop or prop_to_key):
            raise ValueError('at least one of key_to_prop and prop_to_key should be not None')

        if key_to_prop and not callable(key_to_prop):
            TypeError(f"'{type(key_to_prop)}' object is not callable")

        if prop_to_key and not callable(prop_to_key):
            TypeError(f"'{type(prop_to_key)}' object is not callable")

        if flags & Gio.SettingsBindFlags.INVERT_BOOLEAN:
            raise ValueError('Gio.Settings.bind_with_mapping does not support'
                             ' flag Gio.SettingsBindFlags.INVERT_BOOLEAN')

        if not (flags & Gio.SettingsBindFlags.GET or flags & Gio.SettingsBindFlags.SET):
            flags |= Gio.SettingsBindFlags.GET |  Gio.SettingsBindFlags.SET

        if flags & Gio.SettingsBindFlags.GET and key_to_prop:
            obj.notify(prop)
            key_changed(self, key)
            if not (flags & Gio.SettingsBindFlags.GET_NO_CHANGES):
                self.connect('changed::' + key, key_changed)

        if flags & Gio.SettingsBindFlags.SET and prop_to_key:
            obj.connect('notify::' + prop, prop_changed)

        if not (flags & Gio.SettingsBindFlags.NO_SENSITIVITY):
            self.bind_writable(key, obj, "sensitive", False)



from gi.repository import Gtk

from gdms.settings import accessibility_settings, top_bar_settings

from .common import PageContent


class TopBarPageContent (PageContent):
    __gtype_name__ = 'TopBarPageContent'

    def __init__ (self, window, **props):
        super().__init__(**props)

        self.window = window

        self.builder = Gtk.Builder.new_from_resource('/app/ui/top-bar-page.ui')

        self.set_child(self.builder.get_object('content_box'))

        self.text_color_switch = self.builder.get_object('text_color_switch')
        self.text_color_button = self.builder.get_object('text_color_button')
        self.background_color_switch = self.builder.get_object('background_color_switch')
        self.background_color_button = self.builder.get_object('background_color_button')
        self.disable_arrows_row = self.builder.get_object('disable_arrows_row')
        self.disable_rounded_corners_row = self.builder.get_object('disable_rounded_corners_row')
        self.accessibilty_menu_row = self.builder.get_object('accessibilty_menu_row')
        self.show_date_row = self.builder.get_object('show_date_row')
        self.show_weekday_row = self.builder.get_object('show_weekday_row')
        self.show_seconds_row = self.builder.get_object('show_seconds_row')
        self.time_format_comborow = self.builder.get_object('time_format_comborow')
        self.show_battery_percentage_row = self.builder.get_object('show_battery_percentage_row')

        self.bind_to_gsettings()

    def bind_to_gsettings (self):
        top_bar_settings.bind('disable-arrows', self.disable_arrows_row, 'active')
        top_bar_settings.bind('disable-rounded-corners', self.disable_rounded_corners_row, 'active')
        top_bar_settings.bind('change-text-color', self.text_color_switch, 'active')
        top_bar_settings.bind_to_colorbutton('text-color', self.text_color_button)
        top_bar_settings.bind('change-background-color', self.background_color_switch, 'active')
        top_bar_settings.bind_to_colorbutton('background-color', self.background_color_button)
        accessibility_settings.bind('always-show-accessibility-menu',
                                    self.accessibilty_menu_row, 'active')
        top_bar_settings.bind('show-date', self.show_date_row, 'active')
        top_bar_settings.bind('show-weekday', self.show_weekday_row, 'active')
        top_bar_settings.bind('show-seconds', self.show_seconds_row, 'active')
        top_bar_settings.bind_via_list('time-format', self.time_format_comborow, 'selected', ['12h', '24h'])
        top_bar_settings.bind('show-battery-percentage', self.show_battery_percentage_row, 'active')



from gi.repository import Gtk

from gdms.themes import sound_themes
from gdms.settings import sound_settings

from .common import PageContent


class SoundPageContent (PageContent):
    __gtype_name__ = 'SoundPageContent'

    def __init__ (self, window, **props):
        super().__init__(**props)

        self.window = window

        self.builder = Gtk.Builder.new_from_resource('/app/ui/sound-page.ui')

        self.set_child(self.builder.get_object('content_box'))

        self.theme_comborow = self.builder.get_object('theme_comborow')
        self.over_amplification_row = self.builder.get_object('over_amplification_row')
        self.event_sounds_row = self.builder.get_object('event_sounds_row')
        self.feedback_sounds_row = self.builder.get_object('feedback_sounds_row')

        self.theme_comborow.set_model(Gtk.StringList.new(sound_themes.names))
        self.bind_to_gsettings()

    def bind_to_gsettings (self):
        sound_settings.bind_via_list('theme', self.theme_comborow, 'selected',
                                     sound_themes.theme_ids, strict=False)
        sound_settings.bind('event-sounds', self.event_sounds_row, 'active')
        sound_settings.bind('feedback-sounds', self.feedback_sounds_row, 'active')
        sound_settings.bind('over-amplification', self.over_amplification_row, 'active')



import logging
from gettext import gettext as _, pgettext as C_

from gi.repository import Adw
from gi.repository import Gtk

from gdms.utils import BackgroundTask
from gdms.settings import night_light_settings as nl_settings
from gdms import settings

from .common import PageContent

logger = logging.getLogger(__name__)


class DisplayPageContent (PageContent):
    __gtype_name__ = 'DisplayPageContent'

    def __init__ (self, window, **props):
        super().__init__(**props)

        self.window = window

        self.builder = Gtk.Builder.new_from_resource('/app/ui/display-page.ui')

        self.set_child(self.builder.get_object('content_box'))

        self.apply_display_settings_button = self.builder.get_object('apply_display_settings_button')
        self.apply_display_settings_button.connect('clicked', self.on_apply_display_settings)

        self.nl_enable_row = self.builder.get_object('nl_enable_row')
        self.nl_schedule_comborow = self.builder.get_object('nl_schedule_comborow')
        self.nl_start_box = self.builder.get_object('nl_start_box')
        self.nl_start_hour_spinbutton = self.builder.get_object('nl_start_hour_spinbutton')
        self.nl_start_minute_spinbutton = self.builder.get_object('nl_start_minute_spinbutton')
        self.nl_end_box = self.builder.get_object('nl_end_box')
        self.nl_end_hour_spinbutton = self.builder.get_object('nl_end_hour_spinbutton')
        self.nl_end_minute_spinbutton = self.builder.get_object('nl_end_minute_spinbutton')
        self.nl_temperature_scale = self.builder.get_object('nl_temperature_scale')

        self.apply_display_settings_task = BackgroundTask(settings.apply_user_display_settings,
                                                          self.on_apply_display_settings_finish)

        # Following properties are ignored when set in .ui files.
        # So, they need to be changed here.
        self.nl_start_box.set_direction(Gtk.TextDirection.LTR)
        self.nl_end_box.set_direction(Gtk.TextDirection.LTR)
        self.nl_start_hour_spinbutton.set_range(0, 23)
        self.nl_start_hour_spinbutton.set_increments(1,2)
        self.nl_start_minute_spinbutton.set_range(0, 59)
        self.nl_start_minute_spinbutton.set_increments(1,5)
        self.nl_end_hour_spinbutton.set_range(0, 23)
        self.nl_end_hour_spinbutton.set_increments(1,2)
        self.nl_end_minute_spinbutton.set_range(0, 59)
        self.nl_end_minute_spinbutton.set_increments(1,5)
        self.nl_temperature_scale.set_range(1700, 4700)

        self.nl_schedule_comborow.connect('notify::selected', self.update_time_row_sensitivity)

        self.bind_to_gsettings()

    def update_time_row_sensitivity (self, comborow, prop):
        selected = comborow.get_selected()
        self.nl_start_box.set_sensitive(bool(selected))
        self.nl_end_box.set_sensitive(bool(selected))

    def bind_to_gsettings (self):
        nl_settings.bind('enabled', self.nl_enable_row, 'active')
        nl_settings.bind_via_list('schedule-automatic', self.nl_schedule_comborow, 'selected',
                                  [True, False])
        nl_settings.bind('start-hour', self.nl_start_hour_spinbutton, 'value')
        nl_settings.bind('start-minute', self.nl_start_minute_spinbutton, 'value')
        nl_settings.bind('end-hour', self.nl_end_hour_spinbutton, 'value')
        nl_settings.bind('end-minute', self.nl_end_minute_spinbutton, 'value')
        nl_settings.bind('temperature', self.nl_temperature_scale.props.adjustment, 'value')

    def on_apply_display_settings (self, button):
        self.window.task_counter.inc()
        self.apply_display_settings_task.start()

    def on_apply_display_settings_finish(self):
        self.window.task_counter.dec()
        try:
            status = self.apply_display_settings_task.finish()
            if status.success:
                message = _("Applied current display settings")
            else:
                message = _("Failed to apply current display settings")
            toast = Adw.Toast(timeout=2, priority="high", title=message)
            self.window.toast_overlay.add_toast(toast)
        except FileNotFoundError as err:
            logger.debug(f"{err.strerror}: '{err.filename}'")

            message = _("'$XDG_CONFIG_HOME/monitors.xml' file is required to apply current"
                        " display settings but it does not exist.\n"
                        "\n"
                        "In order to create that file automatically, open system 'Settings'"
                        " and change some display options there.")

            dialog = Adw.MessageDialog(
                        body = message,
                       modal = True,
                     heading = _('Monitor Settings Not Found'),
               transient_for = self.window,
             body_use_markup = True,
            )

            dialog.add_response('ok', _('OK'))
            dialog.present()



'''Contains info about themes'''

from os import path
from glob import glob
from gettext import pgettext as C_
from dataclasses import dataclass
import logging

from gi.repository import GLib

from gdms.env import HOST_ROOT, HOST_DATA_DIRS

logger = logging.getLogger(__name__)


@dataclass
class ThemeInfo:
    theme_id: str
    name: str
    path: str

class ThemeList:
    def __init__(self, dirname, decider, add_default=False):
        self.dirname = dirname
        self.decider = decider
        self.add_default = add_default

        self.update()

    def __iter__(self):
        return self._list.__iter__()

    def __getitem__(self, index):
        return self._list[index]

    def __call__(self, theme_id):
        for theme in self:
            if theme.theme_id == theme_id:
                return theme

    def with_name(self, name):
        filtered_list = []
        for theme in self:
            if theme.name == name:
                filtered_list.append(theme)
        return filtered_list

    def _add(self, theme_id, name, path):
        self._list.append(ThemeInfo(theme_id, name, path))

    def _load_index(self, theme_dir):
        filename = path.join(theme_dir, 'index.theme')
        if not path.exists(filename):
            return

        index_file = GLib.KeyFile()
        try:
            index_file.load_from_file(filename,
                                      GLib.KeyFileFlags.KEEP_TRANSLATIONS)
        except Exception:
            logger.debug(f"Failed to parse '{filename}'!")
            return

        return index_file

    def _get_name(self, index_file):
        if not index_file:
            return
        for group in index_file.get_groups()[0]:
            if 'Name' in index_file.get_keys(group)[0]:
                return index_file.get_locale_string(group, 'Name')

    def _is_theme(self, theme_dir, index_file):
        '''checks if @theme_dir@ contains a valid theme of prefered type'''
        return path.exists(path.join(theme_dir, self.decider))

    def update(self):
        self._list = []

        added_ids=[]
        for data_dir in HOST_DATA_DIRS:
            for theme_dir in glob(f"{HOST_ROOT}{data_dir}/{self.dirname}/*"):
                theme_id = path.basename(theme_dir)
                index_file = self._load_index(theme_dir)
                theme_name = self._get_name(index_file) or theme_id
                if theme_id in added_ids or not self._is_theme(theme_dir, index_file):
                    continue
                self._add(theme_id, theme_name, theme_dir)

        self._list.sort(key = lambda x: x.name.lower())

        if self.add_default:
            self._list.insert(0, ThemeInfo('', C_('Theme Name', 'Default'), None))

    @property
    def theme_ids (self):
        id_list = []
        for theme in self:
            id_list.append(theme.theme_id)
        return id_list

    @property
    def names(self):
        name_list = []
        for theme in self:
            name_list.append(theme.name)
        return name_list

    def get_path(self, theme_id):
        if theme := self(theme_id):
            return theme.path

class IconThemeList(ThemeList):
    def __init__(self):
        super().__init__('icons', None)

    def _is_theme(self, theme_dir, index_file):
        if not index_file:
            return False

        if ('Icon Theme' in index_file.get_groups()[0]
        and 'Directories' in index_file.get_keys('Icon Theme')[0]):
            return True

        return False


icon_themes   = IconThemeList()
shell_themes  = ThemeList('themes', 'gnome-shell', True)
sound_themes  = ThemeList('sounds', 'index.theme')
cursor_themes = ThemeList('icons', 'cursors')



import os
from gettext import gettext as _

from gi.repository import Adw, Gtk

from gdms import APP_ID
from gdms.env import TEMP_DIR
from gdms.cmd import CommandList
from gdms.utils import BackgroundTask, GSettings
from gdms import gresource
from gdms import settings

from .common import PageContent


class ToolsPageContent (PageContent):
    __gtype_name__ = 'ToolsPageContent'

    def __init__ (self, window, **props):
        super().__init__(**props)

        self.window = window

        self.builder = Gtk.Builder.new_from_resource('/app/ui/tools-page.ui')

        self.set_child(self.builder.get_object('content_box'))

        self.top_bar_tweaks_row = self.builder.get_object('top_bar_tweaks_row')
        self.extract_shell_theme_button = self.builder.get_object('extract_shell_theme_button')

        self.extract_theme_task = BackgroundTask(self.extract_shell_theme, self.on_extract_shell_theme_finish)
        self.window.task_counter.register(self.extract_shell_theme_button)
        self.extract_shell_theme_button.connect('clicked', self.on_extract_shell_theme)

        # Bind to GSettings
        self.settings = GSettings(APP_ID + '.tools')
        self.settings.bind('top-bar-tweaks', self.top_bar_tweaks_row, 'active')

    def on_extract_shell_theme(self, button):
        self.window.task_counter.inc()
        self.extract_theme_task.start()

    def on_extract_shell_theme_finish(self):
        status, theme_name = self.extract_theme_task.finish()
        self.window.task_counter.dec()

        if status.success:
            # Translators: Do not translate '{folder}' and '{name}'. Keep these as they are.
            # They will be replaced by an actual folder path and theme name during runtime.
            msg = _("Default shell theme extracted to '{folder}' as '{name}'")
            message = msg.format(folder=gresource.ThemesDir, name=theme_name)
        else:
            message = _("Failed to extract default theme")
        toast = Adw.Toast(timeout=2, priority="high", title=message)
        self.window.toast_overlay.add_toast(toast)

    def extract_shell_theme (self):
        perm_theme_name = 'default-extracted'
        perm_theme_path = os.path.join(gresource.ThemesDir, perm_theme_name)
        temp_theme_path = os.path.join(TEMP_DIR, 'extracted-theme')

        # Extract default shell theme to a temporary path
        gresource.extract_default_theme(temp_theme_path)

        # If enabled, apply top bar tweaks
        if self.settings['top-bar-tweaks']:
            with open(os.path.join(temp_theme_path, 'gnome-shell', 'gnome-shell.css'), 'a') as shell_css:
                shell_css.write(settings.get_css())

        # Copy extracted theme to its permanent path
        cmd = CommandList()
        cmd.add('mkdir', '-p', gresource.ThemesDir)
        cmd.add('cp', '-rfT', temp_theme_path, perm_theme_path)
        res = cmd.run(as_root=True)

        return res, perm_theme_name



import os
from gettext import gettext as _, pgettext as C_

from gi.repository import Adw
from gi.repository import Gtk
from gi.repository import GObject

from gdms import APP_NAME, APP_ID, DEBUG
from gdms.utils import BackgroundTask, GProperty, GSettings
from gdms.cmd import Command
from gdms.gresource import UbuntuGdmGresourceFile, BackgroundImageNotFoundError
from gdms import settings

from . import pages
from .sidebar import Sidebar


class TaskCounter(GObject.Object):
    '''A GObject that keeps a count of background tasks and updates widgets accordingly'''

    __gtype_name__ = 'TaskCounter'

    count = GProperty(int, default=0)
    spinner = GProperty(Gtk.Spinner)

    def __init__ (self, **props):
        super().__init__(**props)

        self.widgets = []

        self.connect('notify::count', self.on_count_change)


    @staticmethod
    def on_count_change (self, prop):
        if self.count > 0:
            for widget in self.widgets:
                widget.set_sensitive(False)
            self.spinner.start()
        else:
            for widget in self.widgets:
                widget.set_sensitive(True)
            self.spinner.stop()

    def register (self, widget):
        self.widgets.append(widget)

    def inc (self):
        self.count += 1

    def dec (self):
        self.count -= 1


class GdmSettingsWindow (Adw.ApplicationWindow):
    __gtype_name__ = 'GdmSettingsWindow'

    def __init__ (self, application, **props):
        super().__init__(**props)

        if DEBUG:
            self.add_css_class('devel')

        self.application = application
        self.set_application(application)

        self.props.title = APP_NAME
        self.props.width_request = 360
        self.props.height_request = 360

        self.builder = Gtk.Builder.new_from_resource('/app/ui/main-window.ui')

        self.stack = self.builder.get_object('stack')
        self.sidebar = self.builder.get_object('sidebar')
        self.spinner = self.builder.get_object('spinner')
        self.split_view = self.builder.get_object('split_view')
        self.apply_button = self.builder.get_object('apply_button')
        self.section_label = self.builder.get_object('section_label')
        self.toast_overlay = self.builder.get_object('toast_overlay')

        self.set_content(self.toast_overlay)

        self.task_counter = TaskCounter(spinner=self.spinner)

        self.task_counter.register(self.apply_button)
        self.apply_button.connect('clicked', self.on_apply)
        self.apply_task = BackgroundTask(settings.apply, self.on_apply_finished)

        self.sidebar.connect('activate', self.on_sidebar_activate, self.split_view)
        self.stack.connect('notify::visible-child', self.on_section_changed)

        condition = Adw.BreakpointCondition.parse('max-width: 500sp')
        breakpoint = Adw.Breakpoint.new(condition);
        breakpoint.add_setter(self.split_view, 'collapsed', True)
        self.add_breakpoint(breakpoint);

        self.add_pages()
        self.bind_to_gsettings()

    def on_sidebar_activate (self, sidebar, split_view):
        split_view.props.show_content = True

    def on_section_changed (self, stack, prop):
        current_page = stack.get_page(stack.props.visible_child)
        self.section_label.set_label(current_page.get_title())

    def add_pages (self):

        def add_page(name, title, content):
            page = self.stack.add_titled(content, name, title)
            page.props.icon_name = name + '-symbolic'

        add_page('appearance', _('Appearance'),       pages.AppearancePageContent(self))
        add_page('fonts',      _('Fonts'),            pages.FontsPageContent(self))
        add_page('top_bar',    _('Top Bar'),          pages.TopBarPageContent(self))
        add_page('sound',      _('Sound'),            pages.SoundPageContent(self))
        add_page('pointing',   _('Mouse & Touchpad'), pages.PointingPageContent(self))
        add_page('display',    _('Display'),          pages.DisplayPageContent(self))
        add_page('misc',       _('Login Screen'),     pages.LoginScreenPageContent(self))
        add_page('power',      _('Power'),            pages.PowerPageContent(self))
        add_page('tools',      _('Tools'),            pages.ToolsPageContent(self))

    def bind_to_gsettings (self):
        self.settings = GSettings(APP_ID + '.window-state')

        self.settings.bind('width', self, 'default-width')
        self.settings.bind('height', self, 'default-height')
        self.settings.bind('last-visited-page', self.stack, 'visible-child-name')

    def on_apply (self, button):
        self.task_counter.inc()
        self.apply_task.start()

    def on_apply_finished(self):
        self.task_counter.dec()

        try:
            if self.apply_task.finish():
                message = _('Settings applied successfully')
                if os.environ.get('XDG_CURRENT_DESKTOP') == 'GNOME' and not UbuntuGdmGresourceFile:
                    self.show_logout_dialog()
            else:
                message = _('Failed to apply settings')
            toast = Adw.Toast(timeout=2, priority='high', title=message)

        except BackgroundImageNotFoundError:
            message = _("Didn't apply. Chosen background image does not exist anymore. Please! choose again.")
            toast = Adw.Toast(timeout=4, priority='high', title=message)

        except settings.LogoImageNotFoundError:
            message = _("Didn't apply. Chosen logo image does not exist anymore. Please! choose again.")
            toast = Adw.Toast(timeout=4, priority='high', title=message)

        self.toast_overlay.add_toast(toast)

    def show_logout_dialog (self):
        message = _('The system may start to look weird/buggy until you re-login or reboot.')

        dialog = Adw.MessageDialog(
                    body = message,
                   modal = True,
                 heading = _('Log Out?'),
           transient_for = self,
         body_use_markup = True,
        )

        dialog.add_response('cancel', _('Cancel'))
        dialog.add_response('log-out', _('Log Out'))
        dialog.set_response_appearance('log-out', Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.connect('response', self.on_logout_dialog_response)
        dialog.present()

    def on_logout_dialog_response (self, dialog, response):
        if response == 'log-out':
            Command('gnome-session-quit --no-prompt').run()



from gdms import VERSION
from gdms import env
from gdms import gresource


debug_info = (
    ("Application Version", VERSION),
    ("Operating System", env.OS_PRETTY_NAME),
    ("Package Type", env.PACKAGE_TYPE.name),
    ("TEMP_DIR", env.TEMP_DIR),
    ("HOST_DATA_DIRS", env.HOST_DATA_DIRS),
    ("Main GResource File", gresource.ShellGresourceFile),
    ("Ubuntu GResource File", gresource.UbuntuGdmGresourceFile),
    ("Default GResource File", gresource.get_default()),
)


__markdown_cache = None

def as_markdown():
    global __markdown_cache

    if __markdown_cache:
        return __markdown_cache

    markdown_lines = []
    for key, val in debug_info:
        markdown_lines.append(f"**{key}**: `{val}`  ")

    __markdown_cache = '\n'.join(markdown_lines)
    return __markdown_cache



from gi.repository import Gtk

from gdms.settings import login_screen_settings
from gdms.gui.widgets import ImageChooserButton

from .common import PageContent


class LoginScreenPageContent (PageContent):
    __gtype_name__ = 'LoginScreenPageContent'

    def __init__ (self, window, **props):
        super().__init__(**props)

        self.window = window

        self.builder = Gtk.Builder.new_from_resource('/app/ui/login-screen-page.ui')

        self.set_child(self.builder.get_object('content_box'))

        self.disable_restart_buttons_row = self.builder.get_object('disable_restart_buttons_row')
        self.disable_user_list_row = self.builder.get_object('disable_user_list_row')
        self.enable_fingerprint_authentication_row = (
            self.builder.get_object('enable_fingerprint_authentication_row')
        )
        self.welcome_message_row = self.builder.get_object('welcome_message_row')
        self.welcome_message_entryrow = self.builder.get_object('welcome_message_entryrow')
        self.enlarge_welcome_message_row = self.builder.get_object('enlarge_welcome_message_row')
        self.logo_actionrow = self.builder.get_object('logo_actionrow')
        self.logo_switch = self.builder.get_object('logo_switch')

        # Add logo button
        self.logo_button = ImageChooserButton(valign='center')
        self.logo_actionrow.add_suffix(self.logo_button)
        self.logo_actionrow.set_activatable_widget(self.logo_button)

        self.bind_to_gsettings()

    def bind_to_gsettings (self):
        login_screen_settings.bind('disable-restart-buttons', self.disable_restart_buttons_row, 'active')
        login_screen_settings.bind('disable-user-list', self.disable_user_list_row, 'active')
        login_screen_settings.bind('enable-fingerprint-authentication',
            self.enable_fingerprint_authentication_row, 'active')
        login_screen_settings.bind('enable-welcome-message', self.welcome_message_row, 'active')
        login_screen_settings.bind('enlarge-welcome-message', self.enlarge_welcome_message_row, 'active')
        login_screen_settings.bind('welcome-message', self.welcome_message_entryrow, 'text')
        login_screen_settings.bind('enable-logo', self.logo_switch, 'active')
        login_screen_settings.bind('logo', self.logo_button, 'filename')
